import pytest
from odin.mapping import helpers


class TestMappingHelpers:
    def test_sum_fields(self):
        actual = helpers.sum_fields(22, 20)

        assert actual == 42

    @pytest.mark.parametrize(
        "args, value, expected",
        (
            ((), ["ab", "cd"], "abcd"),
            (("-",), ["ab", "cd"], "ab-cd"),
        ),
    )
    def test_join_fields(self, args, value, expected):
        actual = helpers.join_fields(*args)(*value)

        assert actual == expected

    @pytest.mark.parametrize(
        "args, value, expected",
        (
            ((), "ab cd", ["ab", "cd"]),
            (("-",), "abcd", ["abcd"]),
            (("-",), "ab-cd", ["ab", "cd"]),
            (("-",), "ab-cd-ef", ["ab", "cd", "ef"]),
            (("-", 1), "ab-cd-ef", ["ab", "cd-ef"]),
        ),
    )
    def test_split_field(self, args, value, expected):
        actual = helpers.split_field(*args)(value)

        assert actual == expected



import enum
import uuid
from datetime import datetime
from typing import List, Optional

import odin
from odin.annotated_resource import Options


class Author(odin.AResource):
    name: str

    class Meta:
        namespace = None


class Publisher(odin.AResource):
    name: str

    class Meta:
        namespace = None


class LibraryBook(odin.AResource, abstract=True):
    class Meta:
        namespace = "new_library"


class Book(LibraryBook):
    class Meta:
        key_field_name = "isbn"

    title: str
    isbn: str = odin.Options(
        max_length=32,
    )
    num_pages: Optional[int]
    rrp: float = odin.Options(
        20.4,
        use_default_if_not_provided=True,
    )
    fiction: bool = True
    genre: str = Options(
        choices=(
            ("sci-fi", "Science Fiction"),
            ("fantasy", "Fantasy"),
            ("biography", "Biography"),
            ("others", "Others"),
            ("computers-and-tech", "Computers & technology"),
        ),
    )
    published: List[datetime]
    authors: List[Author] = odin.Options(
        use_container=True,
    )
    publisher: Optional[Publisher]

    A_CONST = "Foo Bar"

    def __eq__(self, other):
        if other:
            return vars(self) == vars(other)
        return False


class From(enum.Enum):
    Dumpster = "dumpster"
    Shop = "shop"
    Ebay = "ebay"


class IdentifiableBook(Book):
    id: uuid.UUID
    purchased_from: From


class Subscriber(odin.Resource):
    """
    Mixed resource types
    """

    name: str = odin.StringField()
    address: str = odin.StringField()


class Library(odin.AnnotatedResource):
    name: str
    books: List[LibraryBook]
    subscribers: Optional[Subscriber]

    @odin.calculated_field
    def book_count(self):
        return len(self.books)

    class Meta:
        namespace = "new_library"



import pathlib
import re

import pytest
import datetime
import enum
import uuid
from copy import deepcopy

import odin
from odin.fields import *
from odin.fields import Field, TimeStampField, NotProvided
from odin.datetimeutil import FixedTimezone
from odin.fields.virtual import VirtualField
from odin.validators import (
    MinValueValidator,
    MaxValueValidator,
    MaxLengthValidator,
    RegexValidator,
    MinLengthValidator,
)
from odin.exceptions import ValidationError

utc = datetime.timezone.utc


class ObjectValue:
    pass


class ValidatorTest:
    message = "Default message"
    code = "test_code"

    def __init__(self, fail, *params):
        self.fail = fail
        self.params = params

    def __call__(self, value):
        if self.fail:
            raise ValidationError(
                code=self.code, message=self.message, params=self.params
            )


class FieldTest(Field):
    def to_python(self, value):
        return value


class DynamicTypeNameFieldTest(IntegerField):
    @staticmethod
    def data_type_name(instance):
        return "Foo"


class TestField:
    def test_error_messages_no_overrides(self):
        target = FieldTest()

        assert {
            "invalid_choice": "Value %r is not a valid choice.",
            "null": "This field cannot be null.",
            "required": "This field is required.",
        } == target.error_messages

    def test_error_messages_override_add(self):
        target = FieldTest(
            error_messages={
                "null": "Override",
                "other": "Other Value",
            }
        )

        assert {
            "invalid_choice": "Value %r is not a valid choice.",
            "null": "Override",
            "required": "This field is required.",
            "other": "Other Value",
        } == target.error_messages

    def test_set_attributes_from_name(self):
        target = FieldTest()
        target.set_attributes_from_name("test_name")

        assert "test_name" == target.name
        assert "test_name" == target.attname
        assert "test name" == target.verbose_name
        assert "test names" == target.verbose_name_plural

    def test_set_attributes_from_name_with_name(self):
        target = FieldTest(name="init_name")
        target.set_attributes_from_name("test_name")

        assert "init_name" == target.name
        assert "test_name" == target.attname
        assert "init name" == target.verbose_name
        assert "init names" == target.verbose_name_plural

    def test_set_attributes_from_name_with_verbose_name(self):
        target = FieldTest(verbose_name="init Verbose Name")
        target.set_attributes_from_name("test_name")

        assert "test_name" == target.name
        assert "test_name" == target.attname
        assert "init Verbose Name" == target.verbose_name
        assert "init Verbose Names" == target.verbose_name_plural

    @pytest.mark.parametrize(
        "value, expected",
        (
            (None, None),
            ("", ""),
            ("abc", "abc"),
            (1, "1"),
            (True, "True"),
            (False, "False"),
        ),
    )
    def test_as_string(self, value, expected):
        target = FieldTest()

        actual = target.as_string(value)

        assert actual == expected

    def test_choice_doc_test(self):
        target = FieldTest(
            choices=(
                ("a", "A Value"),
                ("b", "B Value"),
                ("c", "C Value"),
                ("d", "D Value"),
            )
        )

        assert target.choices_doc_text == target.choices

    def test_has_default(self):
        target = FieldTest()

        assert not target.has_default()

    def test_has_default_supplied(self):
        target = FieldTest(default="123")

        assert target.has_default()

    def test_get_default(self):
        target = FieldTest()

        assert target.get_default() is None

    def test_get_default_supplied(self):
        target = FieldTest(default="123")

        assert "123" == target.get_default()

    def test_get_default_callable(self):
        target = FieldTest(default=lambda: "321")

        assert "321" == target.get_default()

    def test_value_from_object(self):
        target = FieldTest()
        target.set_attributes_from_name("test_name")

        an_obj = ObjectValue()
        setattr(an_obj, "test_name", "test_value")

        actual = target.value_from_object(an_obj)
        assert "test_value" == actual

    def test__repr(self):
        target = Field()
        assert "<odin.fields.Field>" == repr(target)
        target.set_attributes_from_name("eek")
        assert "<odin.fields.Field: eek>" == repr(target)

    def test__deep_copy(self):
        field = FieldTest(name="Test")
        target_copy = deepcopy(field)
        target_assign = field
        assert field is target_assign
        assert field is not target_copy
        assert field.name == target_copy.name

    def test_run_validators_and_override_validator_message(self):
        target = FieldTest(
            error_messages={"test_code": "Override message"},
            validators=[ValidatorTest(True)],
        )

        try:
            target.run_validators("Placeholder")
        except ValidationError as ve:
            assert "Override message" == ve.messages[0]
        else:
            raise AssertionError("Validation Error not raised.")

    def test_run_validators_and_override_validator_message_with_params(self):
        target = FieldTest(
            error_messages={"test_code": "Override message: %s"},
            validators=[ValidatorTest(True, "123")],
        )

        try:
            target.run_validators("Placeholder")
        except ValidationError as ve:
            assert ["Override message: 123"] == ve.messages
        else:
            raise AssertionError("Validation Error not raised.")

    def test_default_to_python_raises_not_implemented(self):
        target = Field()
        pytest.raises(NotImplementedError, target.to_python, "Anything...")

    def test_clean_uses_default_if_value_is_not_provided_is_true(self):
        target = FieldTest(use_default_if_not_provided=True, default="foo")
        actual = target.clean(NotProvided)
        assert "foo" == actual

    def test_clean_uses_default_if_value_is_not_provided_is_false(self):
        # Need to allow None as the if use_default_if_not_provided is false NOT_PROVIDED evaluates to None.
        target = FieldTest(use_default_if_not_provided=False, default="foo", null=True)
        actual = target.clean(NotProvided)
        assert actual is None


class VirtualFieldTest(VirtualField):
    pass


class TestVirtualField:
    def test_creation_counter(self):
        current_count = VirtualField.creation_counter
        next_count = current_count + 1
        target = VirtualField()

        assert VirtualField.creation_counter == next_count
        assert target.creation_counter == current_count
        assert hash(target) == current_count

    def test_repr(self):
        target = VirtualFieldTest()
        assert "<tests.test_fields.VirtualFieldTest>" == repr(target)
        target.set_attributes_from_name("eek")
        assert "<tests.test_fields.VirtualFieldTest: eek>" == repr(target)

    def test_default_descriptor_behaviour(self):
        class TestObj:
            test_field = VirtualField()

        target = TestObj()

        with pytest.raises(NotImplementedError):
            _ = target.test_field

        with pytest.raises(AttributeError) as excinfo:
            target.test_field = 123

        assert "Read only" == str(excinfo.value)


class TestFields:
    @staticmethod
    def assert_validator_in(validator_class, validators):
        """
        Assert that the specified validator is in the validation list.
        :param validator_class:
        :param validators:
        """
        for v in validators:
            if isinstance(v, validator_class):
                return
        raise AssertionError(
            "Validator %r was not found in list of validators." % validator_class
        )

    @staticmethod
    def assert_validator_not_in(validator_class, validators):
        """
        Assert that the specified validator is not in the validation list.
        :param validator_class:
        :param validators:
        """
        for v in validators:
            if isinstance(v, validator_class):
                raise AssertionError(
                    "Validator %r was found in list of validators." % validator_class
                )

    # BooleanField ############################################################

    @pytest.mark.parametrize(
        ("field", "value", "actual"),
        (
            (BooleanField(), True, True),
            (BooleanField(), 1, True),
            (BooleanField(), "Yes", True),
            (BooleanField(), "true", True),
            (BooleanField(), "T", True),
            (BooleanField(), "1", True),
            (BooleanField(), "TRUE", True),
            (BooleanField(), False, False),
            (BooleanField(), 0, False),
            (BooleanField(), "No", False),
            (BooleanField(), "false", False),
            (BooleanField(), "FALSE", False),
            (BooleanField(), "F", False),
            (BooleanField(), "0", False),
            (BooleanField(null=True), None, None),
        ),
    )
    def test_booleanfield_success(self, field, value, actual):
        assert field.clean(value) == actual

    @pytest.mark.parametrize(
        ("field", "value"),
        (
            (BooleanField(), None),
            (BooleanField(), ""),
            (BooleanField(), "Awesome!"),
        ),
    )
    def test_booleanfield_failure(self, field, value):
        with pytest.raises(ValidationError):
            field.clean(value)

    # StringField #############################################################

    @pytest.mark.parametrize(
        ("field", "value", "actual"),
        (
            (StringField(), "1", "1"),
            (StringField(), "eek", "eek"),
            (StringField(null=True), "1", "1"),
            (StringField(null=True), None, None),
            (StringField(null=True), "", ""),
            (StringField(null=True, empty=True), "", ""),
            (StringField(empty=True), "", ""),
            (StringField(max_length=10), "123456", "123456"),
        ),
    )
    def test_stringfield_success(self, field, value, actual):
        assert field.clean(value) == actual

    @pytest.mark.parametrize(
        ("field", "value"),
        (
            (StringField(), None),
            (StringField(), ""),
            (StringField(null=True, empty=False), ""),
            (StringField(empty=False), ""),
            (StringField(max_length=10), "1234567890a"),
        ),
    )
    def test_stringfield_failure(self, field, value):
        with pytest.raises(ValidationError):
            field.clean(value)

    def test_stringfield(self):
        f = StringField()
        assert f.max_length is None
        self.assert_validator_not_in(MaxLengthValidator, f.validators)

        f = StringField(max_length=10)
        assert f.max_length == 10
        self.assert_validator_in(MaxLengthValidator, f.validators)

    @pytest.mark.parametrize(
        ("field", "empty_value"),
        (
            (StringField(), False),
            (StringField(null=True), True),
            (StringField(null=False), False),
            (StringField(null=True, empty=True), True),
            (StringField(null=False, empty=True), True),
            (StringField(null=True, empty=False), False),
            (StringField(null=False, empty=False), False),
        ),
    )
    def test_stringfield__handling_of_null_empty(self, field, empty_value):
        assert field.empty == empty_value

    # URLField ################################################################

    def test_urlfield_1(self):
        f = UrlField()
        assert "http://www.github.com" == f.clean("http://www.github.com")
        pytest.raises(ValidationError, f.clean, "eek")
        pytest.raises(ValidationError, f.clean, None)
        assert f.max_length == None
        self.assert_validator_in(RegexValidator, f.validators)

    # IntegerField ############################################################

    def test_integerfield_1(self):
        f = IntegerField()
        pytest.raises(ValidationError, f.clean, None)
        pytest.raises(ValidationError, f.clean, "abc")
        assert 123 == f.clean(123)
        assert 123 == f.clean("123")
        assert 123 == f.clean(123.5)
        assert None == f.min_value
        self.assert_validator_not_in(MinValueValidator, f.validators)
        assert None == f.max_value
        self.assert_validator_not_in(MaxValueValidator, f.validators)

    def test_integerfield_2(self):
        f = IntegerField(null=True)
        assert None == f.clean(None)
        pytest.raises(ValidationError, f.clean, "abc")
        assert 69 == f.clean(69)
        assert 69 == f.clean("69")
        assert 69 == f.clean(69.5)
        assert None == f.min_value
        self.assert_validator_not_in(MinValueValidator, f.validators)
        assert None == f.max_value
        self.assert_validator_not_in(MaxValueValidator, f.validators)

    def test_integerfield_3(self):
        f = IntegerField(min_value=50, max_value=100)
        pytest.raises(ValidationError, f.clean, None)
        pytest.raises(ValidationError, f.clean, "abc")
        assert 69 == f.clean(69)
        assert 69 == f.clean("69")
        assert 69 == f.clean(69.5)
        assert 50 == f.clean(50)
        assert 100 == f.clean(100)
        pytest.raises(ValidationError, f.clean, 30)
        pytest.raises(ValidationError, f.clean, 110)
        assert 50 == f.min_value
        self.assert_validator_in(MinValueValidator, f.validators)
        assert 100 == f.max_value
        self.assert_validator_in(MaxValueValidator, f.validators)

    # FloatField ##############################################################

    def test_floatfield_1(self):
        f = FloatField()
        pytest.raises(ValidationError, f.clean, None)
        pytest.raises(ValidationError, f.clean, "abc")
        assert 123 == f.clean(123)
        assert 123.5 == f.clean("123.5")
        assert 123.5 == f.clean(123.5)
        assert None == f.min_value
        self.assert_validator_not_in(MinValueValidator, f.validators)
        assert None == f.max_value
        self.assert_validator_not_in(MaxValueValidator, f.validators)

    def test_floatfield_2(self):
        f = FloatField(null=True)
        assert f.clean(None) is None
        pytest.raises(ValidationError, f.clean, "abc")
        assert 69 == f.clean(69)
        assert 69.5 == f.clean("69.5")
        assert 69.5 == f.clean(69.5)
        assert None == f.min_value
        self.assert_validator_not_in(MinValueValidator, f.validators)
        assert None == f.max_value
        self.assert_validator_not_in(MaxValueValidator, f.validators)

    def test_floatfield_3(self):
        f = FloatField(min_value=50.5, max_value=100.4)
        pytest.raises(ValidationError, f.clean, None)
        pytest.raises(ValidationError, f.clean, "abc")
        assert 69 == f.clean(69)
        assert 69.5 == f.clean("69.5")
        assert 69.5 == f.clean(69.5)
        assert 50.5 == f.clean(50.5)
        assert 100.4 == f.clean(100.4)
        pytest.raises(ValidationError, f.clean, 30)
        pytest.raises(ValidationError, f.clean, 110)
        assert 50.5 == f.min_value
        self.assert_validator_in(MinValueValidator, f.validators)
        assert 100.4 == f.max_value
        self.assert_validator_in(MaxValueValidator, f.validators)

    # DateField ###############################################################

    @pytest.mark.parametrize(
        "field,value,expected".split(","),
        (
            (DateField(), "2013-11-24", datetime.date(2013, 11, 24)),
            (DateField(), datetime.date(2013, 11, 24), datetime.date(2013, 11, 24)),
            (
                DateField(),
                datetime.datetime(2013, 11, 24, 1, 14),
                datetime.date(2013, 11, 24),
            ),
            (DateField(null=True), None, None),
        ),
    )
    def test_datefield_clean_success(self, field, value, expected):
        assert field.clean(value) == expected

    @pytest.mark.parametrize(
        "field,value".split(","),
        (
            (DateField(), None),
            (DateField(), "abc"),
            (DateField(), 123),
        ),
    )
    def test_datefield_clean_failure(self, field, value):
        pytest.raises(ValidationError, field.clean, value)

    @pytest.mark.parametrize(
        "field,value,expected".split(","),
        ((DateField(), datetime.date(2013, 11, 24), "2013-11-24"),),
    )
    def test_datefield_as_string(self, field, value, expected):
        assert field.as_string(value) == expected

    # TimeField ###############################################################

    def test_timefield_1(self):
        f = TimeField(assume_local=False)
        pytest.raises(ValidationError, f.clean, None)
        pytest.raises(ValidationError, f.clean, "abc")
        pytest.raises(ValidationError, f.clean, 123)
        assert datetime.time(18, 43, tzinfo=utc) == f.clean("18:43:00.000Z")
        assert datetime.time(18, 43, tzinfo=utc) == f.clean(
            datetime.time(18, 43, tzinfo=utc)
        )

    def test_timefield_2(self):
        f = TimeField(assume_local=False, null=True)
        assert f.clean(None) is None
        pytest.raises(ValidationError, f.clean, "abc")
        pytest.raises(ValidationError, f.clean, 123)
        assert datetime.time(18, 43, tzinfo=utc) == f.clean("18:43:00.000Z")
        assert datetime.time(18, 43, tzinfo=utc) == f.clean(
            datetime.time(18, 43, tzinfo=utc)
        )

    @pytest.mark.parametrize(
        "field,value,expected".split(","),
        (
            (TimeField(), datetime.time(12, 44, 12, 12), "12:44:12.000012"),
            (TimeField(), datetime.time(12, 44, 12, 12, utc), "12:44:12.000012+00:00"),
        ),
    )
    def test_timefield_as_string(self, field, value, expected):
        assert field.as_string(value) == expected

    # NaiveTimeField ##########################################################

    @pytest.mark.parametrize(
        ("target", "value", "expected"),
        (
            (
                NaiveTimeField(ignore_timezone=False),
                "18:43:00.000Z",
                datetime.time(18, 43, tzinfo=utc),
            ),
            (
                NaiveTimeField(ignore_timezone=False),
                "18:43:00.000Z",
                datetime.time(18, 43, tzinfo=utc),
            ),
            (
                NaiveTimeField(ignore_timezone=False),
                datetime.time(18, 43, tzinfo=utc),
                datetime.time(18, 43, tzinfo=utc),
            ),
            (NaiveTimeField(ignore_timezone=False, null=True), None, None),
            (
                NaiveTimeField(ignore_timezone=False, null=True),
                "18:43:00.000Z",
                datetime.time(18, 43, tzinfo=utc),
            ),
            (
                NaiveTimeField(ignore_timezone=False, null=True),
                "18:43:00.000Z",
                datetime.time(18, 43, tzinfo=utc),
            ),
            (
                NaiveTimeField(ignore_timezone=False, null=True),
                datetime.time(18, 43, tzinfo=utc),
                datetime.time(18, 43, tzinfo=utc),
            ),
            (
                NaiveTimeField(ignore_timezone=True),
                "18:43:00.000Z",
                datetime.time(18, 43),
            ),
            (
                NaiveTimeField(ignore_timezone=True),
                "18:43:00.000Z",
                datetime.time(18, 43),
            ),
            (
                NaiveTimeField(ignore_timezone=True),
                datetime.time(18, 43, tzinfo=utc),
                datetime.time(18, 43),
            ),
        ),
    )
    def test_naivetimefield__clean_valid_values(self, target, value, expected):
        assert target.clean(value) == expected

    @pytest.mark.parametrize(
        ("target", "value"),
        (
            (NaiveTimeField(ignore_timezone=False), None),
            (NaiveTimeField(ignore_timezone=False), "abc"),
            (NaiveTimeField(ignore_timezone=False), 123),
            (NaiveTimeField(ignore_timezone=False, null=True), "abc"),
            (NaiveTimeField(ignore_timezone=False, null=True), 123),
            (NaiveTimeField(ignore_timezone=True), None),
            (NaiveTimeField(ignore_timezone=True), "abc"),
            (NaiveTimeField(ignore_timezone=True), 123),
        ),
    )
    def test_naivetimefield__clean_invalid_values(self, target, value):
        pytest.raises(ValidationError, target.clean, value)

    @pytest.mark.parametrize(
        ("target", "value", "expected"),
        (
            (
                NaiveTimeField(ignore_timezone=False),
                datetime.time(18, 43, tzinfo=utc),
                datetime.time(18, 43, tzinfo=utc),
            ),
            (
                NaiveTimeField(ignore_timezone=True),
                datetime.time(18, 43, tzinfo=utc),
                datetime.time(18, 43),
            ),
        ),
    )
    def test_naivetimefield__prepare(self, target, value, expected):
        assert target.prepare(value) == expected

    # DateTimeField ###########################################################

    def test_datetimefield_1(self):
        f = DateTimeField(assume_local=False)
        pytest.raises(ValidationError, f.clean, None)
        pytest.raises(ValidationError, f.clean, "abc")
        pytest.raises(ValidationError, f.clean, 123)
        assert datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc) == f.clean(
            "2013-11-24T18:43:00.000Z"
        )
        assert datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc) == f.clean(
            "2013-11-24 18:43:00.000Z"
        )
        assert datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc) == f.clean(
            datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc)
        )

    def test_datetimefield_2(self):
        f = DateTimeField(assume_local=False, null=True)
        assert f.clean(None) is None
        pytest.raises(ValidationError, f.clean, "abc")
        pytest.raises(ValidationError, f.clean, 123)
        assert datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc) == f.clean(
            "2013-11-24T18:43:00.000Z"
        )
        assert datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc) == f.clean(
            "2013-11-24 18:43:00.000Z"
        )
        assert datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc) == f.clean(
            datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc)
        )

    @pytest.mark.parametrize(
        "field,value,expected".split(","),
        (
            (
                DateTimeField(),
                datetime.datetime(
                    2013, 11, 24, 18, 43, tzinfo=FixedTimezone.from_hours_minutes(10)
                ),
                "2013-11-24T18:43:00+10:00",
            ),
            (
                DateTimeField(),
                datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc),
                "2013-11-24T18:43:00+00:00",
            ),
        ),
    )
    def test_datetimefield_as_string(self, field, value, expected):
        assert field.as_string(value) == expected

    # NaiveDateTimeField ######################################################

    @pytest.mark.parametrize(
        ("target", "value", "expected"),
        (
            (
                NaiveDateTimeField(ignore_timezone=False),
                "2013-11-24T18:43:00.000Z",
                datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc),
            ),
            (
                NaiveDateTimeField(ignore_timezone=False),
                "2013-11-24 18:43:00.000Z",
                datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc),
            ),
            (
                NaiveDateTimeField(ignore_timezone=False),
                datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc),
                datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc),
            ),
            (NaiveDateTimeField(ignore_timezone=False, null=True), None, None),
            (
                NaiveDateTimeField(ignore_timezone=False, null=True),
                "2013-11-24T18:43:00.000Z",
                datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc),
            ),
            (
                NaiveDateTimeField(ignore_timezone=False, null=True),
                "2013-11-24 18:43:00.000Z",
                datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc),
            ),
            (
                NaiveDateTimeField(ignore_timezone=False, null=True),
                datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc),
                datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc),
            ),
            (
                NaiveDateTimeField(ignore_timezone=True),
                "2013-11-24T18:43:00.000Z",
                datetime.datetime(2013, 11, 24, 18, 43),
            ),
            (
                NaiveDateTimeField(ignore_timezone=True),
                "2013-11-24 18:43:00.000Z",
                datetime.datetime(2013, 11, 24, 18, 43),
            ),
            (
                NaiveDateTimeField(ignore_timezone=True),
                datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc),
                datetime.datetime(2013, 11, 24, 18, 43),
            ),
        ),
    )
    def test_naivedatetimefield__clean_valid_values(self, target, value, expected):
        assert target.clean(value) == expected

    @pytest.mark.parametrize(
        ("target", "value"),
        (
            (NaiveDateTimeField(ignore_timezone=False), None),
            (NaiveDateTimeField(ignore_timezone=False), "abc"),
            (NaiveDateTimeField(ignore_timezone=False), 123),
            (NaiveDateTimeField(ignore_timezone=False, null=True), "abc"),
            (NaiveDateTimeField(ignore_timezone=False, null=True), 123),
            (NaiveDateTimeField(ignore_timezone=True), None),
            (NaiveDateTimeField(ignore_timezone=True), "abc"),
            (NaiveDateTimeField(ignore_timezone=True), 123),
        ),
    )
    def test_naivedatetimefield__clean_invalid_values(self, target, value):
        pytest.raises(ValidationError, target.clean, value)

    @pytest.mark.parametrize(
        ("target", "value", "expected"),
        (
            (
                NaiveDateTimeField(ignore_timezone=False),
                datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc),
                datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc),
            ),
            (
                NaiveDateTimeField(ignore_timezone=True),
                datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc),
                datetime.datetime(2013, 11, 24, 18, 43),
            ),
        ),
    )
    def test_naivedatetimefield__prepare(self, target, value, expected):
        assert target.prepare(value) == expected

    # HttpDateTimeField #######################################################

    def test_httpdatetimefield_1(self):
        f = HttpDateTimeField()
        pytest.raises(ValidationError, f.clean, None)
        pytest.raises(ValidationError, f.clean, "abc")
        pytest.raises(ValidationError, f.clean, 123)
        assert datetime.datetime(2012, 8, 29, 17, 12, 58, tzinfo=utc) == f.clean(
            "Wed Aug 29 17:12:58 +0000 2012"
        )
        assert datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc) == f.clean(
            datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc)
        )

    def test_httpdatetimefield_2(self):
        f = HttpDateTimeField(null=True)
        assert f.clean(None) is None
        pytest.raises(ValidationError, f.clean, "abc")
        pytest.raises(ValidationError, f.clean, 123)
        assert datetime.datetime(2012, 8, 29, 17, 12, 58, tzinfo=utc) == f.clean(
            "Wed Aug 29 17:12:58 +0000 2012"
        )
        assert datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc) == f.clean(
            datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc)
        )

    def test_httpdate_time_field__as_string(self):
        target = HttpDateTimeField()

        actual = target.as_string(datetime.datetime(2019, 6, 29, 8, 45, 23, tzinfo=utc))

        assert actual == "Sat, 29 Jun 2019 08:45:23 GMT"

    # TimeStampField ##########################################################

    def test_timestampfield_1(self):
        f = TimeStampField()
        pytest.raises(ValidationError, f.clean, None)
        pytest.raises(ValidationError, f.clean, "abc")
        pytest.raises(ValidationError, f.clean, "Wed Aug 29 17:12:58 +0000 2012")
        assert datetime.datetime(1970, 1, 1, 0, 2, 3, tzinfo=utc) == f.clean(123)
        assert datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc) == f.clean(
            datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc)
        )

    def test_timestampfield_2(self):
        f = TimeStampField(null=True)
        assert f.clean(None) is None
        pytest.raises(ValidationError, f.clean, "abc")
        pytest.raises(ValidationError, f.clean, "Wed Aug 29 17:12:58 +0000 2012")
        assert datetime.datetime(1970, 1, 1, 0, 2, 3, tzinfo=utc) == f.clean(123)
        assert datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc) == f.clean(
            datetime.datetime(2013, 11, 24, 18, 43, tzinfo=utc)
        )

    def test_timestampfield_3(self):
        f = TimeStampField()
        assert f.prepare(None) is None
        assert 123 == f.prepare(datetime.datetime(1970, 1, 1, 0, 2, 3, tzinfo=utc))
        assert 123 == f.prepare(123)
        assert 123 == f.prepare(
            datetime.datetime(
                1970, 1, 1, 10, 2, 3, tzinfo=FixedTimezone.from_hours_minutes(10)
            )
        )

    # DictField ###############################################################

    def test_dictfield__where_null_is_false(self):
        f = DictField()
        pytest.raises(ValidationError, f.clean, None)
        pytest.raises(ValidationError, f.clean, "abc")
        pytest.raises(ValidationError, f.clean, 123)
        assert {} == f.clean({})
        assert {"foo": "bar"} == f.clean({"foo": "bar"})
        assert f.default == dict

    def test_dictfield__where_null_is_true(self):
        f = DictField(null=True)
        assert None is f.clean(None)
        assert {} == f.clean({})
        pytest.raises(ValidationError, f.clean, "abc")
        pytest.raises(ValidationError, f.clean, 123)
        assert {"foo": "bar"} == f.clean({"foo": "bar"})

    def test_dictfield__validators_are_executed_on_empty(self):
        """
        Ensure that validators are executed even if the field is empty
        """
        f = DictField(validators=[MinLengthValidator(1)])
        pytest.raises(ValidationError, f.clean, {})

    # ListField ##############################################################

    def test_listfield__where_null_is_false(self):
        f = ListField()
        pytest.raises(ValidationError, f.clean, None)
        pytest.raises(ValidationError, f.clean, "abc")
        pytest.raises(ValidationError, f.clean, 123)
        assert [] == f.clean([])
        assert ["foo", "bar"], f.clean(["foo" == "bar"])
        assert ["foo", "bar", "$", "eek"], f.clean(["foo", "bar", "$" == "eek"])
        assert f.default == list

    def test_listfield__where_null_is_true(self):
        f = ListField(null=True)
        assert None == f.clean(None)
        pytest.raises(ValidationError, f.clean, "abc")
        pytest.raises(ValidationError, f.clean, 123)
        assert [] == f.clean([])
        assert ["foo", "bar"], f.clean(["foo" == "bar"])
        assert ["foo", "bar", "$", "eek"], f.clean(["foo", "bar", "$" == "eek"])

    def test_listfield__validators_are_executed_on_empty(self):
        """
        Ensure that validators are executed even if the field is empty
        """
        f = ListField(validators=[MinLengthValidator(1)])
        pytest.raises(ValidationError, f.clean, [])

    # TypedListField #########################################################

    def test_typedlistfield__where_null_is_false(self):
        f = TypedListField(IntegerField())
        assert "List<Integer>" == f.data_type_name(f)
        pytest.raises(ValidationError, f.clean, None)
        pytest.raises(ValidationError, f.clean, "abc")
        pytest.raises(ValidationError, f.clean, 123)
        assert [] == f.clean([])
        pytest.raises(ValidationError, f.clean, ["foo", "bar"])
        assert [1, 2, 3], f.clean([1, 2 == 3])
        assert f.default == list

    def test_typedlistfield__where_null_is_true(self):
        f = TypedListField(IntegerField(), null=True)
        assert "List<Integer>" == f.data_type_name(f)
        assert f.clean(None) is None
        pytest.raises(ValidationError, f.clean, "abc")
        pytest.raises(ValidationError, f.clean, 123)
        assert [] == f.clean([])
        pytest.raises(ValidationError, f.clean, ["foo", "bar"])
        assert [1, 2, 3], f.clean([1, 2 == 3])

    def test_typed_list_field__dynamic_type_name(self):
        f = TypedListField(DynamicTypeNameFieldTest(), null=True)
        assert "List<Foo>" == f.data_type_name(f)

    @pytest.mark.parametrize(
        "target, expected",
        (
            (TypedListField(IntegerField()), None),
            (
                TypedListField(
                    IntegerField(),
                    choices=[("these", "1"), ("are", "2"), ("bad", "3")],
                ),
                [("these", "1"), ("are", "2"), ("bad", "3")],
            ),
            (
                TypedListField(IntegerField(choices=((1, "1"), (2, "2"), (3, "3")))),
                ((1, "1"), (2, "2"), (3, "3")),
            ),
            (
                TypedListField(
                    IntegerField(choices=[("these", "1"), ("are", "2"), ("bad", "3")]),
                ),
                [("these", "1"), ("are", "2"), ("bad", "3")],
            ),
        ),
    )
    def test_typed_list_field__with_choices(self, target, expected):
        assert target.choices_doc_text == expected

    @pytest.mark.parametrize("value", ([None], [10, 11, 3], ["Fifteen"]))
    def test_typed_list_field__with_nested_validation_errors(self, value):
        target = TypedListField(IntegerField(min_value=10, null=False))
        pytest.raises(ValidationError, target.clean, value)

    # TypedDictField ##########################################################

    def test_typeddictfield_1(self):
        f = TypedDictField(IntegerField())
        assert "Dict<String, Integer>" == f.data_type_name(f)
        pytest.raises(ValidationError, f.clean, None)
        pytest.raises(ValidationError, f.clean, "abc")
        pytest.raises(ValidationError, f.clean, 123)
        assert {} == f.clean({})
        pytest.raises(ValidationError, f.clean, {"foo": "bar"})
        assert {"foo": 1} == f.clean({"foo": 1})

    def test_typeddictfield_2(self):
        f = TypedDictField(IntegerField(), null=True)
        assert "Dict<String, Integer>" == f.data_type_name(f)
        assert None == f.clean(None)
        pytest.raises(ValidationError, f.clean, "abc")
        pytest.raises(ValidationError, f.clean, 123)
        assert {} == f.clean({})
        pytest.raises(ValidationError, f.clean, {"foo": "bar"})
        assert {"foo": 1} == f.clean({"foo": 1})

    def test_typeddictfield_3(self):
        f = TypedDictField(StringField(), IntegerField(), null=True)
        assert "Dict<Integer, String>" == f.data_type_name(f)
        pytest.raises(ValidationError, f.clean, {"foo": "bar"})
        assert {1: "foo"} == f.clean({1: "foo"})

    def test_typeddictfield_nested_typed_array(self):
        f = TypedDictField(TypedArrayField(StringField()))
        assert "Dict<String, List<String>>" == f.data_type_name(f)
        assert {} == f.clean({})
        pytest.raises(ValidationError, f.clean, {"foo": "bar"})
        assert {"foo": ["bar", "eek"]}, f.clean({"foo": ["bar" == "eek"]})

    def test_typeddictfield_validate(self):
        f = TypedDictField(
            IntegerField(min_value=5),
            StringField(
                max_length=5,
                choices=[
                    ("foo", "Foo"),
                    ("bad_value", "Bad Value"),
                ],
            ),
        )
        assert "Dict<String, Integer>" == f.data_type_name(f)
        pytest.raises(ValidationError, f.clean, {None: 6})
        pytest.raises(ValidationError, f.clean, {"bad_value": 6})
        pytest.raises(ValidationError, f.clean, {"bar": 6})
        pytest.raises(ValidationError, f.clean, {"foo": None})
        pytest.raises(ValidationError, f.clean, {"foo": 2})

    def test_typed_dict_field_dynamic_type_name(self):
        f = TypedDictField(
            DynamicTypeNameFieldTest(),
            DynamicTypeNameFieldTest(),
        )
        assert "Dict<Foo, Foo>" == f.data_type_name(f)

    @pytest.mark.parametrize(
        "field value actual".split(),
        (
            (EmailField(), "foo@example.company", "foo@example.company"),
            (IPv4Field(), "127.0.0.1", "127.0.0.1"),
            (IPv6Field(), "::1", "::1"),
            (IPv6Field(), "1:2:3:4:5:6:7:8", "1:2:3:4:5:6:7:8"),
            (IPv46Field(), "127.0.0.1", "127.0.0.1"),
            (IPv46Field(), "::1", "::1"),
            (IPv46Field(), "1:2:3:4:5:6:7:8", "1:2:3:4:5:6:7:8"),
        ),
    )
    def test_valid_values(self, field, value, actual):
        assert field.clean(value) == actual

    # UUIDField ################################################################

    @pytest.mark.parametrize(
        "value",
        (
            uuid.uuid1(),
            uuid.uuid3(uuid.uuid4(), "name"),
            uuid.uuid4(),
            uuid.uuid5(uuid.uuid4(), "name"),
        ),
    )
    def test_uuid_field_with_uuid_objects(self, value):
        f = UUIDField()

        assert f.clean(value) == value

    @pytest.mark.parametrize(
        "value",
        (
            uuid.uuid1().bytes,
            uuid.uuid3(uuid.uuid4(), "name").bytes,
            uuid.uuid4().bytes,
            uuid.uuid5(uuid.uuid4(), "name").bytes,
        ),
        ids=(
            "bytes-uuid1",
            "bytes-uuid3",
            "bytes-uuid4",
            "bytes-uuid5",
        ),
    )
    def test_uuid_field_with_16_bytes_sequence(self, value):
        f = UUIDField()

        assert f.clean(value) == uuid.UUID(bytes=value)

    @pytest.mark.parametrize(
        "value",
        (
            str(uuid.uuid1()).encode("utf-8"),
            str(uuid.uuid3(uuid.uuid4(), "name")).encode("utf-8"),
            str(uuid.uuid4()).encode("utf-8"),
            str(uuid.uuid5(uuid.uuid4(), "name")).encode("utf-8"),
        ),
    )
    def test_uuid_field_with_bytes(self, value):
        f = UUIDField()

        assert f.clean(value) == uuid.UUID(value.decode("utf-8"))

    @pytest.mark.parametrize(
        "value",
        (
            str(uuid.uuid1()),
            str(uuid.uuid3(uuid.uuid4(), "name")),
            str(uuid.uuid4()),
            str(uuid.uuid5(uuid.uuid4(), "name")),
        ),
    )
    def test_uuid_field_with_string(self, value):
        f = UUIDField()

        assert f.clean(value) == uuid.UUID(value)

    @pytest.mark.parametrize("value", range(4))
    def test_uuid_field_with_int(self, value):
        f = UUIDField()

        assert f.clean(value) == uuid.UUID(int=value)

    @pytest.mark.parametrize("value", (-1, -2))
    def test_uuid_field_invalid_int(self, value):
        f = UUIDField()
        with pytest.raises(ValidationError):
            f.clean(value)

    @pytest.mark.parametrize(
        "value",
        (
            b"\254",
            b"\255",
        ),
    )
    def test_uuid_field_invalid_bytes(self, value):
        f = UUIDField()
        with pytest.raises(ValidationError):
            f.clean(value)

    @pytest.mark.parametrize(
        "value",
        (
            (1, 2, 3, 4, 5),
            [1, 2, 3, 4, 5],
            (-1, 2, 2, 2, 2, 2),
            [-1, 2, 2, 2, 2, 2],
            (2, -1, 2, 2, 2, 2),
            [2, -1, 2, 2, 2, 2],
        ),
    )
    def test_uuid_field_invalid_fields(self, value):
        f = UUIDField()
        with pytest.raises(ValidationError):
            f.clean(value)

    @pytest.mark.parametrize(
        "value", ("sometext", "01010101-0101-0101-0101-01010101010")
    )
    def test_uuid_field_invalid_hex(self, value):
        f = UUIDField()
        with pytest.raises(ValidationError):
            f.clean(value)

    def test_uuid_field_non_str_value(self):
        some_uuid = uuid.uuid4()

        class SomeObject:
            def __str__(self):
                return str(some_uuid)

        f = UUIDField()

        assert f.clean(SomeObject()) == some_uuid

    def test_uuid_field_invalid_non_str_value(self):
        class SomeObject:
            def __str__(self):
                return "sometext"

        f = UUIDField()

        with pytest.raises(ValidationError):
            f.clean(SomeObject())

    def test_uuid_field__none(self):
        f = UUIDField(null=True)

        assert f.clean(None) is None

    def test_uuid_field__not_none(self):
        f = UUIDField()

        with pytest.raises(ValidationError):
            f.clean(None)


class Colour(enum.Enum):
    Red = "red"
    Green = "green"
    Blue = "blue"


class Options(enum.Enum):
    Undefined = ""
    Yes = "True"
    No = "False"


class Number(enum.IntEnum):
    Thirteen = 13
    FortyTwo = 42
    SixtyNine = 69


class TestEnumField:
    @pytest.mark.parametrize(
        ("field", "value", "actual"),
        (
            (EnumField(Colour), "red", Colour.Red),
            (EnumField(Colour), "green", Colour.Green),
            (EnumField(Colour), Colour.Blue, Colour.Blue),
            (EnumField(Colour, null=True), None, None),
            (EnumField(Colour, null=True), "", None),
            (EnumField(Options), "", Options.Undefined),
            (EnumField(Number), 13, Number.Thirteen),
            (EnumField(Number), 42, Number.FortyTwo),
            (EnumField(Number), Number.SixtyNine, Number.SixtyNine),
            (EnumField(Number, null=True), None, None),
        ),
    )
    def test_to_python__valid(self, field, value, actual):
        assert field.clean(value) == actual

    @pytest.mark.parametrize(
        ("field", "value"),
        (
            (EnumField(Colour), "pink"),
            (EnumField(Colour), True),
            (EnumField(Colour), None),
            (EnumField(Colour), ""),
            (EnumField(Colour), Number.Thirteen),
            (EnumField(Number), "13"),
            (EnumField(Number), 14),
            (EnumField(Number), None),
        ),
    )
    def test_to_python__invalid(self, field, value):
        with pytest.raises(ValidationError):
            field.clean(value)

    @pytest.mark.parametrize(
        "field, value, actual",
        (
            (EnumField(Colour), "red", None),
            (EnumField(Colour), Colour.Blue, "blue"),
            (EnumField(Colour), None, None),
            (EnumField(Number), 13, None),
            (EnumField(Number), Number.FortyTwo, 42),
            (EnumField(Number), None, None),
        ),
    )
    def test_prepare(self, field, value, actual):
        assert field.prepare(value) == actual

    @pytest.mark.parametrize(
        "field, expected",
        (
            (
                EnumField(Colour),
                ((Colour.Red, "Red"), (Colour.Green, "Green"), (Colour.Blue, "Blue")),
            ),
            (
                EnumField(Number),
                (
                    (Number.Thirteen, "Thirteen"),
                    (Number.FortyTwo, "FortyTwo"),
                    (Number.SixtyNine, "SixtyNine"),
                ),
            ),
            (
                EnumField(Colour, choices=(Colour.Red, Colour.Blue)),
                ((Colour.Red, "Red"), (Colour.Blue, "Blue")),
            ),
        ),
    )
    @pytest.mark.skipif("sys.version_info.major < 3")
    def test_choices(self, field, expected):
        assert field.choices == expected

    @pytest.mark.parametrize(
        "field, expected",
        (
            (
                EnumField(Colour),
                {(Colour.Red, "Red"), (Colour.Green, "Green"), (Colour.Blue, "Blue")},
            ),
            (
                EnumField(Number),
                {
                    (Number.Thirteen, "Thirteen"),
                    (Number.FortyTwo, "FortyTwo"),
                    (Number.SixtyNine, "SixtyNine"),
                },
            ),
            (
                EnumField(Colour, choices=(Colour.Red, Colour.Blue)),
                {(Colour.Red, "Red"), (Colour.Blue, "Blue")},
            ),
        ),
    )
    def test_choices__with_set(self, field, expected):
        """Use sets to handle random ordering in earlier Python releases"""
        assert set(field.choices) == expected

    @pytest.mark.parametrize(
        "field, expected",
        (
            (EnumField(Colour), (("red", "Red"), ("green", "Green"), ("blue", "Blue"))),
            (
                EnumField(Number),
                ((13, "Thirteen"), (42, "FortyTwo"), (69, "SixtyNine")),
            ),
            (
                EnumField(Colour, choices=(Colour.Red, Colour.Blue)),
                (("red", "Red"), ("blue", "Blue")),
            ),
        ),
    )
    @pytest.mark.skipif("sys.version_info.major < 3")
    def test_choices_doc_text(self, field, expected):
        assert field.choices_doc_text == expected

    @pytest.mark.parametrize(
        "field, expected",
        (
            (EnumField(Colour), {("red", "Red"), ("green", "Green"), ("blue", "Blue")}),
            (
                EnumField(Number),
                {(13, "Thirteen"), (42, "FortyTwo"), (69, "SixtyNine")},
            ),
            (
                EnumField(Colour, choices=(Colour.Red, Colour.Blue)),
                {("red", "Red"), ("blue", "Blue")},
            ),
        ),
    )
    def test_choices_doc_text__with_sets(self, field, expected):
        """Use sets to handle random ordering in earlier Python releases"""
        assert set(field.choices_doc_text) == expected


class TestPathField:
    @pytest.mark.parametrize(
        "value, expected",
        (
            (None, None),
            ("abc", pathlib.Path("abc")),
            ("/this/is/the/true/path", pathlib.Path("/this/is/the/true/path")),
        ),
    )
    def test_to_python__valid(self, value, expected):
        field = odin.PathField()

        actual = field.to_python(value)

        assert actual == expected

    @pytest.mark.parametrize("value", (123,))
    def test_to_python__invalid(self, value):
        field = odin.PathField()

        with pytest.raises(ValidationError):
            field.to_python(value)

    @pytest.mark.parametrize(
        "value, expected",
        (
            (None, None),
            ("abc", None),
            (pathlib.Path("abc"), "abc"),
            (pathlib.Path("/this/is/the/true/path"), "/this/is/the/true/path"),
        ),
    )
    def test_prepare(self, value, expected):
        field = odin.PathField()

        actual = field.prepare(value)

        assert actual == expected


class TestRegexField:
    @pytest.mark.parametrize(
        "value, expected",
        (
            (None, None),
            ("simple[0-9](match)", re.compile("simple[0-9](match)")),
        ),
    )
    def test_to_python__valid(self, value, expected):
        field = odin.RegexField()

        actual = field.to_python(value)

        assert actual == expected

    @pytest.mark.parametrize("value", (123, "invalid[0-9(expression)"))
    def test_to_python__invalid(self, value):
        field = odin.RegexField()

        with pytest.raises(ValidationError):
            field.to_python(value)

    @pytest.mark.parametrize(
        "value, expected",
        (
            (None, None),
            ("simple[0-9](match)", None),
            (re.compile("simple[0-9](match)"), "simple[0-9](match)"),
        ),
    )
    def test_prepare(self, value, expected):
        field = odin.RegexField()

        actual = field.prepare(value)

        assert actual == expected



from typing import Union, Sequence, Dict, Optional, Type

from odin import resources, ResourceAdapter
from odin.exceptions import CodecDecodeError
from odin.resources import ResourceBase
from odin.utils import getmeta

try:
    import toml
except ImportError:  # pragma: no cover
    raise ImportError(  # pragma: no cover
        "odin.codecs.toml_codec requires the 'toml' package."
    )


TOML_TYPES = {}
CONTENT_TYPE = "application/toml"


def load(fp, resource=None, full_clean=True, default_to_not_supplied=False):
    """
    Load a resource from a TOML encoded file.

    If a ``resource`` value is supplied it is used as the base resource for the supplied YAML. If one is not supplied a
    resource type field ``$`` is used to obtain the type represented by the dictionary. A ``ValidationError`` will be
    raised if either of these values are supplied and not compatible. It is valid for a type to be supplied in the file
    to be a child object from within the inheritance tree.

    :param fp: a file pointer to read TOML data format.
    :param resource: A resource type, resource name or list of resources and names to use as the base for creating a
        resource. If a list is supplied the first item will be used if a resource type is not supplied.
    :param full_clean: Do a full clean of the object as part of the loading process.
    :param default_to_not_supplied: Used for loading partial resources. Any fields not supplied are replaced with
        NOT_SUPPLIED.
    :returns: A resource object or object graph of resources loaded from file.

    """

    try:
        data = toml.load(fp)
    except toml.TomlDecodeError as ex:
        raise CodecDecodeError(str(ex))
    return resources.build_object_graph(
        data,
        resource,
        full_clean,
        False,
        default_to_not_supplied,
    )


def loads(s, resource=None, full_clean=True, default_to_not_supplied=False):
    """Load a resource from a TOML encoded string.

    If a ``resource`` value is supplied it is used as the base resource for the supplied YAML. If one is not supplied a
    resource type field ``$`` is used to obtain the type represented by the dictionary. A ``ValidationError`` will be
    raised if either of these values are supplied and not compatible. It is valid for a type to be supplied in the file
    to be a child object from within the inheritance tree.

    :param s: a string containing TOML.
    :param resource: A resource type, resource name or list of resources and names to use as the base for creating a
        resource. If a list is supplied the first item will be used if a resource type is not supplied.
    :param full_clean: Do a full clean of the object as part of the loading process.
    :param default_to_not_supplied: Used for loading partial resources. Any fields not supplied are replaced with
        NOT_SUPPLIED.
    :returns: A resource object or object graph of resources loaded from file.

    """

    try:
        data = toml.loads(s)
    except toml.TomlDecodeError as ex:
        raise CodecDecodeError(str(ex))
    return resources.build_object_graph(
        data,
        resource,
        full_clean,
        False,
        default_to_not_supplied,
    )


class OdinEncoder(toml.TomlEncoder):
    """Encode that handles Odin types."""

    def __init__(
        self,
        include_virtual_fields: bool = True,
        include_type_field: bool = True,
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.include_virtual_fields = include_virtual_fields
        self.include_type_field = include_type_field

    def resource_to_dict(self, v):
        resource_dict = v.to_dict(self.include_virtual_fields)
        if self.include_type_field:
            meta = getmeta(v)
            resource_dict[meta.type_field] = meta.resource_name
        return resource_dict

    def dump_value(self, v):
        if isinstance(v, (ResourceBase, ResourceAdapter)):
            resource_dict = self.resource_to_dict(v)
            return self.dump_inline_table(resource_dict)
        if type(v) in TOML_TYPES:
            v = TOML_TYPES[type(v)](v)
        return super().dump_value(v)


RT = Union[
    ResourceBase,
    ResourceAdapter,
    Sequence[ResourceBase],
    Sequence[ResourceAdapter],
    Dict,
]


def dump(
    resource: RT,
    fp,
    encoder: Optional[Type[OdinEncoder]] = None,
    include_virtual_fields: bool = True,
    **kwargs
):
    """Dump to a TOML encoded file.

    :param resource: The root resource to dump to a JSON encoded file.
    :param fp: The file pointer that represents the output file.
    :param encoder: Encoder to use serializing to a string; default is the :py:class:`OdinEncoder`.
    :param include_virtual_fields: Include virtual fields in the output
    :param kwargs: Additional keyword arguments for the encoder.
    """
    encoder = (encoder or OdinEncoder)(include_virtual_fields, **kwargs)

    if isinstance(resource, (ResourceBase, ResourceAdapter)):
        resource = encoder.resource_to_dict(resource)

    toml.dump(resource, fp, encoder)


def dumps(
    resource: RT,
    encoder: Optional[Type[OdinEncoder]] = None,
    include_virtual_fields: bool = True,
    **kwargs
) -> str:
    """Dump to a TOML encoded file.

    :param resource: The root resource to dump to a JSON encoded file.
    :param encoder: Encoder to use serializing to a string; default is the :py:class:`OdinEncoder`.
    :param include_virtual_fields: Include virtual fields in the output
    :param kwargs: Additional keyword arguments for the encoder.
    :returns: TOML encoded string.
    """
    encoder = (encoder or OdinEncoder)(include_virtual_fields, **kwargs)

    if isinstance(resource, (ResourceBase, ResourceAdapter)):
        resource = encoder.resource_to_dict(resource)

    return toml.dumps(resource, encoder)



import pytest
import datetime

from odin import datetimeutil


class TestTimezones:
    def test_fixed_timezone(self):
        target = datetimeutil.FixedTimezone.from_hours_minutes(10, 30)

        assert datetime.timedelta(hours=10, minutes=30) == target.utcoffset(
            datetime.datetime.now()
        )
        assert "10:30" == target.tzname(datetime.datetime.now())
        assert datetime.timedelta(0) == target.dst(datetime.datetime.now())
        assert "10:30" == str(target)

    def test_fixed_timezone_pickling(self):
        import pickle

        target = datetimeutil.FixedTimezone.from_hours_minutes(10, 30)

        data = pickle.dumps(target)
        result = pickle.loads(data)

        assert result == target


class TestParseIsoDateString:
    @pytest.mark.parametrize("value", ("2014-01-13", "2014-1-13"))
    def test_valid_values(self, value):
        assert datetime.date(2014, 1, 13) == datetimeutil.parse_iso_date_string(value)

    @pytest.mark.parametrize("value", (123, "2014/01/13"))
    def test_invalid_values(self, value):
        pytest.raises(ValueError, datetimeutil.parse_iso_date_string, value)


class TestParseIsoTimeString:
    @pytest.mark.parametrize(
        ("value", "expected"),
        (
            ("23:53:25", datetime.time(23, 53, 25, 0, datetimeutil.utc)),
            ("23:53:25Z", datetime.time(23, 53, 25, 0, datetimeutil.utc)),
            ("23:53:25.432Z", datetime.time(23, 53, 25, 432, datetimeutil.utc)),
            (
                "23:53:25+10",
                datetime.time(
                    23, 53, 25, 0, datetimeutil.FixedTimezone.from_hours_minutes(10, 0)
                ),
            ),
            (
                "23:53:25.432+10",
                datetime.time(
                    23,
                    53,
                    25,
                    432,
                    datetimeutil.FixedTimezone.from_hours_minutes(10, 0),
                ),
            ),
            (
                "23:53:25+10:30",
                datetime.time(
                    23, 53, 25, 0, datetimeutil.FixedTimezone.from_hours_minutes(10, 30)
                ),
            ),
            (
                "23:53:25.432-10:30",
                datetime.time(
                    23,
                    53,
                    25,
                    432,
                    datetimeutil.FixedTimezone.from_hours_minutes(-10, -30),
                ),
            ),
        ),
    )
    def test_valid_values(self, value, expected):
        assert datetimeutil.parse_iso_time_string(value) == expected

    @pytest.mark.parametrize(
        ("value", "default_tz", "expected"),
        (
            ("23:53:25", None, datetime.time(23, 53, 25)),
            ("23:53:25", datetimeutil.IgnoreTimezone, datetime.time(23, 53, 25)),
            ("23:53:25Z", None, datetime.time(23, 53, 25, 0, datetimeutil.utc)),
            ("23:53:25Z", datetimeutil.IgnoreTimezone, datetime.time(23, 53, 25, 0)),
        ),
    )
    def test_valid_values_timezone_handling(self, value, default_tz, expected):
        assert datetimeutil.parse_iso_time_string(value, default_tz) == expected

    def test_value_values_with_alternate_timezone(self):
        assert datetime.time(
            23, 53, 25, 0, datetimeutil.FixedTimezone.from_hours_minutes(10, 0)
        ) == datetimeutil.parse_iso_time_string(
            "23:53:25",
            default_timezone=datetimeutil.FixedTimezone.from_hours_minutes(10, 0),
        )

    @pytest.mark.parametrize("value", (123, "23:53:25EST"))
    def test_invalid_values(self, value):
        pytest.raises(ValueError, datetimeutil.parse_iso_time_string, value)


class TestParseIsoDateTimeString:
    @pytest.mark.parametrize(
        ("value", "expected"),
        (
            (
                "2014-01-13T00:28:33",
                datetime.datetime(2014, 1, 13, 0, 28, 33, 0, datetimeutil.utc),
            ),
            (
                "2014-01-13T00:28:33Z",
                datetime.datetime(2014, 1, 13, 0, 28, 33, 0, datetimeutil.utc),
            ),
            (
                "2014-01-13T00:28:33.432Z",
                datetime.datetime(2014, 1, 13, 0, 28, 33, 432, datetimeutil.utc),
            ),
            (
                "2014-01-13T00:28:33+10",
                datetime.datetime(
                    2014,
                    1,
                    13,
                    0,
                    28,
                    33,
                    0,
                    datetimeutil.FixedTimezone.from_hours_minutes(10, 0),
                ),
            ),
            (
                "2014-01-13T00:28:33.432+10",
                datetime.datetime(
                    2014,
                    1,
                    13,
                    0,
                    28,
                    33,
                    432,
                    datetimeutil.FixedTimezone.from_hours_minutes(10, 0),
                ),
            ),
            (
                "2014-01-13T00:28:33+10:30",
                datetime.datetime(
                    2014,
                    1,
                    13,
                    0,
                    28,
                    33,
                    0,
                    datetimeutil.FixedTimezone.from_hours_minutes(10, 30),
                ),
            ),
            (
                "2014-01-13T00:28:33.432-10:30",
                datetime.datetime(
                    2014,
                    1,
                    13,
                    0,
                    28,
                    33,
                    432,
                    datetimeutil.FixedTimezone.from_hours_minutes(-10, -30),
                ),
            ),
            (
                "2014-01-13T00:28:33.432+1030",
                datetime.datetime(
                    2014,
                    1,
                    13,
                    0,
                    28,
                    33,
                    432,
                    datetimeutil.FixedTimezone.from_hours_minutes(10, 30),
                ),
            ),
            # Non-Standard but close formats that are common
            (
                "2014-01-13 00:28:33Z",
                datetime.datetime(2014, 1, 13, 0, 28, 33, 0, datetimeutil.utc),
            ),
            (
                "2014-01-13 00:28:33 GMT",
                datetime.datetime(2014, 1, 13, 0, 28, 33, 0, datetimeutil.utc),
            ),
        ),
    )
    def test_valid_values(self, value, expected):
        assert datetimeutil.parse_iso_datetime_string(value) == expected

    @pytest.mark.parametrize(
        ("value", "default_tz", "expected"),
        (
            ("2014-01-13T00:28:33", None, datetime.datetime(2014, 1, 13, 0, 28, 33)),
            (
                "2014-01-13T00:28:33",
                datetimeutil.IgnoreTimezone,
                datetime.datetime(2014, 1, 13, 0, 28, 33),
            ),
            (
                "2014-01-13T00:28:33Z",
                None,
                datetime.datetime(2014, 1, 13, 0, 28, 33, 0, datetimeutil.utc),
            ),
            (
                "2014-01-13T00:28:33Z",
                datetimeutil.IgnoreTimezone,
                datetime.datetime(2014, 1, 13, 0, 28, 33),
            ),
        ),
    )
    def test_valid_values_timezone_handling(self, value, default_tz, expected):
        assert datetimeutil.parse_iso_datetime_string(value, default_tz) == expected

    @pytest.mark.parametrize(
        "value",
        (
            123,
            "2014-1-13T00:28:33Z",
            "2014/1/13T00:28:33Z",
            "2014-01-13T00:28:33EST",
        ),
    )
    def test_invalid_values(self, value):
        pytest.raises(ValueError, datetimeutil.parse_iso_datetime_string, value)


class TestToDateString:
    def test_naive_datetime(self):
        dt = datetime.datetime(2013, 7, 13, 16, 54, 46, 123000)
        actual = datetimeutil.to_ecma_datetime_string(
            dt, datetimeutil.utc
        )  # assume UTC to simplify this particular test
        assert "2013-07-13T16:54:46.123Z" == actual

    def test_aware_datetime(self):
        dt = datetime.datetime(2013, 7, 13, 16, 54, 46, 123000, datetimeutil.utc)
        actual = datetimeutil.to_ecma_datetime_string(dt, datetimeutil.utc)
        assert "2013-07-13T16:54:46.123Z" == actual


class TestParseHttpDateString:
    @pytest.mark.parametrize(
        ("value", "expected"),
        (
            (
                "Wed Aug 29 17:12:58 +1030 2012",
                datetime.datetime(
                    2012,
                    8,
                    29,
                    17,
                    12,
                    58,
                    0,
                    datetimeutil.FixedTimezone.from_hours_minutes(10, 30),
                ),
            ),
            (
                "Wed Aug 29 17:12:58 -0900 2012",
                datetime.datetime(
                    2012,
                    8,
                    29,
                    17,
                    12,
                    58,
                    0,
                    datetimeutil.FixedTimezone.from_hours_minutes(-9, 0),
                ),
            ),
            (
                "Wed Aug 29 17:12:58 -0930 2012",
                datetime.datetime(
                    2012,
                    8,
                    29,
                    17,
                    12,
                    58,
                    0,
                    datetimeutil.FixedTimezone.from_hours_minutes(-9, -30),
                ),
            ),
            (
                "Wed Aug 29 17:12:58 +0800 2012",
                datetime.datetime(
                    2012,
                    8,
                    29,
                    17,
                    12,
                    58,
                    0,
                    datetimeutil.FixedTimezone.from_hours_minutes(8, 00),
                ),
            ),
            # Alternate position for the year
            (
                "Wed, Aug 29 2012 17:12:58 -0930",
                datetime.datetime(
                    2012,
                    8,
                    29,
                    17,
                    12,
                    58,
                    0,
                    datetimeutil.FixedTimezone.from_hours_minutes(-9, -30),
                ),
            ),
        ),
    )
    def test_valid_values(self, value, expected):
        assert datetimeutil.parse_http_datetime_string(value) == expected

    @pytest.mark.parametrize(
        "value",
        (
            123,
            "2014-1-13T00:28:33Z",
            "2014/1/13T00:28:33Z",
            "2014-01-13T00:28:33EST",
        ),
    )
    def test_invalid_values(self, value):
        pytest.raises(ValueError, datetimeutil.parse_http_datetime_string, value)


class TestToHttpDateString:
    @pytest.mark.parametrize(
        ("value", "expected"),
        (
            (
                datetime.datetime(
                    2012,
                    8,
                    29,
                    17,
                    12,
                    58,
                    0,
                    datetimeutil.FixedTimezone.from_hours_minutes(10, 30),
                ),
                "Wed, 29 Aug 2012 06:42:58 GMT",
            ),
            (
                datetime.datetime(2012, 8, 29, 17, 12, 58, 0, datetimeutil.utc),
                "Wed, 29 Aug 2012 17:12:58 GMT",
            ),
        ),
    )
    def test_valid_values(self, value, expected):
        assert datetimeutil.to_http_datetime_string(value) == expected



import pytest

from odin import datastructures


class TestCaseLessStringList:
    @pytest.mark.parametrize(
        "needle haystack".split(),
        (
            ("foo", ["foo"]),
            ("FOO", ["foo"]),
            ("foo", ["FOO"]),
            ("Foo", ["fOO"]),
            ("Foo", ["foo", "bar"]),
        ),
    )
    def test_contains__true(self, needle, haystack):
        target = datastructures.CaseLessStringList(haystack)
        assert needle in target

    @pytest.mark.parametrize(
        "needle haystack".split(),
        (
            ("foo", ["food"]),
            ("FOO", ["food"]),
            ("foo", ["FOOD"]),
            ("Foo", ["fOOd"]),
            ("Foo", ["food", "bar"]),
        ),
    )
    def test_contains__false(self, needle, haystack):
        target = datastructures.CaseLessStringList(haystack)
        assert needle not in target

    def test_extend(self):
        target = datastructures.CaseLessStringList(("foo",))
        target.extend("bar")
        assert target == ["FOO", "B", "A", "R"]

    def test_append(self):
        target = datastructures.CaseLessStringList("foo")
        target.append("bar")
        assert "Bar" in target

    def test_insert(self):
        target = datastructures.CaseLessStringList(("foo", "bar"))
        target.insert(1, "eek")
        assert "Eek" in target
        assert target == ["FOO", "EEK", "BAR"]

    def test_index(self):
        target = datastructures.CaseLessStringList(("foo", "bar", "eek"))

        assert target.index("Foo") == 0
        assert target.index("Bar") == 1
        assert target.index("EEk", 1) == 2

    def test_remove(self):
        target = datastructures.CaseLessStringList(("foo", "bar", "eek"))
        target.remove("Bar")
        assert target == ["FOO", "EEK"]



import pytest

from odin import Resource
from odin.contrib import sphinx


class TestResourceDocumenter:
    @pytest.mark.parametrize(
        "type_, expected",
        (
            (1, False),
            ("EEk", False),
            (Resource, True),
        ),
    )
    def test_can_document_member(self, type_, expected):
        actual = sphinx.ResourceDocumenter.can_document_member(type_)

        assert actual == expected



import abc
import operator
from typing import Callable, Iterable, Any, Sequence, Union

from .exceptions import InvalidPathError
from .resources import ResourceBase
from .traversal import TraversalPath


class FilterAtom(abc.ABC):
    """
    Base filter statement
    """

    __slots__ = ()

    @abc.abstractmethod
    def __call__(self, resource: ResourceBase) -> bool:
        """
        Call filter with a resource
        """

    def any(self, collection):
        return any(self(r) for r in collection)

    def all(self, collection):
        return all(self(r) for r in collection)


class FilterChain(FilterAtom):
    """
    Collection of filter atoms to match resources
    """

    __slots__ = ("_atoms",)

    operator_name = ""
    check_operator: Callable[[Iterable[bool]], bool] = all

    def __init__(self, *atoms):
        self._atoms = list(atoms)

    def __len__(self):
        return len(self._atoms)

    def __add__(self, other):
        if isinstance(other, self.__class__):
            return self.__class__(*self._atoms + other._atoms)
        elif isinstance(other, FilterComparison):
            self._atoms.append(other)
            return self
        raise TypeError(f"{other} not supported for this operation")

    def __call__(self, resource: ResourceBase) -> bool:
        return self.check_operator(a(resource) for a in self._atoms)

    def __str__(self):
        if not self._atoms:
            return ""
        pin = f" {self.operator_name} "
        return f"({pin.join(str(a) for a in self._atoms)})"


class And(FilterChain):
    """
    Chain of filters we all must match; logic AND
    """

    __slots__ = ()

    operator_name = "AND"
    check_operator = all


class Or(FilterChain):
    """
    Chain of filters we any can match; logic OR
    """

    __slots__ = ()

    operator_name = "OR"
    check_operator = any


class FilterComparison(FilterAtom, abc.ABC):
    """
    Base class for filter operator atoms
    """

    __slots__ = ("field", "value", "operation")

    # Symbol for this operator and alternatives. The first item is used when generating
    # a representation of the filter, the others are used for parsing queries.
    operator_symbols: Sequence[str] = []
    compare_operator: Callable[[Any, Any], bool]

    def __init__(
        self,
        field: Union[TraversalPath, str],
        value: Any,
        operation: Callable[[Any], Any] = None,
    ):
        self.field = TraversalPath.parse(field)
        self.value = value
        self.operation = operation

    def __call__(self, resource: ResourceBase) -> bool:
        """
        Call to the filter to compare a resource
        """
        try:
            value = self.field.get_value(resource)
        except InvalidPathError:
            return False
        else:
            if self.operation:
                value = self.operation(value)
            return self.compare_operator(value, self.value)

    def __str__(self):
        value = self.value
        if isinstance(self.value, str):
            value = f"{value!r}"

        if self.operation:
            op_name = getattr(self.operation, "name", self.operation.__name__)
            return f"{op_name}({self.field}) {self.operator_symbols[0]} {value}"
        else:
            return f"{self.field} {self.operator_symbols[0]} {value}"


class Equal(FilterComparison):
    """
    Equal filter operator
    """

    __slots__ = ()

    operator_symbols = ("==", "=", "eq")
    compare_operator = operator.eq


class NotEqual(FilterComparison):
    """
    Not equal filter operator
    """

    __slots__ = ()

    operator_symbols = ("!=", "<>", "neq")
    compare_operator = operator.ne


class LessThan(FilterComparison):
    """
    Less than filter operator
    """

    __slots__ = ()

    operator_symbols = ("<", "lt")
    compare_operator = operator.lt


class LessThanOrEqual(FilterComparison):
    """
    Less than or equal filter operator
    """

    __slots__ = ()

    operator_symbols = ("<=", "lte")
    compare_operator = operator.le


class GreaterThan(FilterComparison):
    """
    Greater than filter operator
    """

    __slots__ = ()

    operator_symbols = (">", "gt")
    compare_operator = operator.gt


class GreaterThanOrEqual(FilterComparison):
    """
    Greater than or equal filter operator
    """

    __slots__ = ()

    operator_symbols = (">=", "gte")
    compare_operator = operator.ge



"""Methods for resolving types to fields."""
import datetime
import enum
import pathlib
import re
import uuid
from typing import Any, Sequence, Dict, Type, Union, get_origin, List, Final

try:
    # Handle the change in typing between 3.8 and later releases
    from typing import T, KT, VT
except ImportError:
    T = None
    KT = None
    VT = None

import odin
from .special_fields import AnyField
from .type_aliases import Validator, Choices, Email, IPv4, IPv6, IPv46, Url
from .. import ListOf, DictOf
from ..exceptions import ResourceDefError
from ..fields import (
    BaseField,
    Field,
    NotProvided,
    TypedListField,
    TypedDictField,
    ListField,
    DictField,
)
from ..fields.virtual import ConstantField
from ..resources import ResourceBase


class Options:
    """Define options for a field"""

    __slots__ = (
        "field_type",
        "base_args",
        "field_args",
        "extra_args",
    )

    def __init__(
        self,
        default: Any = odin.NotProvided,
        field_type: Type[odin.BaseField] = None,
        *,
        verbose_name: str = None,
        verbose_name_plural: str = None,
        name: str = None,
        validators: Sequence[Validator] = None,
        choices: Choices = None,
        use_default_if_not_provided: bool = True,
        error_messages: Dict[str, str] = None,
        doc_text: str = "",
        is_attribute: bool = False,
        key: bool = False,
        **extra_args: Any,
    ):
        self.field_type = field_type
        self.base_args = {
            "verbose_name": verbose_name,
            "verbose_name_plural": verbose_name_plural,
            "name": name,
            "doc_text": doc_text,
            "is_attribute": is_attribute,
            "key": key,
        }
        self.base_args.update(extra_args)
        self.field_args = {
            "choices": choices,
            "use_default_if_not_provided": use_default_if_not_provided,
            "default": default,
            "validators": validators,
            "error_messages": error_messages,
        }

    def _kwargs(self):
        """Build kwargs used to instantiate field"""
        final_args = self.base_args
        if issubclass(self.field_type, Field):
            final_args.update(self.field_args)
        return final_args

    def _improve_error(self, ex: TypeError):
        """Attempt to provide more context for the error"""
        message = str(ex)

        for check in (
            "Field.__init__()",
            "VirtualField.__init__()",
            f"{self.field_type.__name__}.__init__()",
        ):
            if message.startswith(check):
                raise ResourceDefError(
                    f"{self.field_type.__name__}.__init__(){message[len(check):]}"
                )

    def init_field(self):
        """Instantiate field object"""
        if self.field_type:
            try:
                return self.field_type(**self._kwargs())
            except TypeError as ex:
                self._improve_error(ex)
                raise
        raise ResourceDefError(
            f"Field type `{self.base_args.get('name', 'Unknown!')}` could not be resolved"
        )


SIMPLE_TYPE_MAP = {
    bool: odin.BooleanField,
    datetime.date: odin.DateField,
    datetime.datetime: odin.DateTimeField,
    dict: odin.DictField,
    Email: odin.EmailField,
    float: odin.FloatField,
    int: odin.IntegerField,
    IPv4: odin.IPv4Field,
    IPv6: odin.IPv6Field,
    IPv46: odin.IPv46Field,
    list: odin.ListField,
    pathlib.Path: odin.PathField,
    re.Pattern: odin.RegexField,
    str: odin.StringField,
    datetime.time: odin.TimeField,
    Url: odin.UrlField,
    uuid.UUID: odin.UUIDField,
    Any: AnyField,  # For Python 3.11
}


def _resolve_field_from_type(options: Options, type_: type):
    """Resolve a field from a basic type"""
    if field_type := SIMPLE_TYPE_MAP.get(type_, None):
        options.field_type = field_type

    # Is a resource; resolve as DictAs
    elif issubclass(type_, ResourceBase):
        options.field_args["resource"] = type_
        options.field_type = odin.DictAs

    # Is an enum; resolve as EnumType
    elif issubclass(type_, enum.Enum):
        options.field_args["enum_type"] = type_
        options.field_type = odin.EnumField

    else:
        raise ResourceDefError(f"Unable to resolve field for type {type_}")


def is_optional(type_: Union) -> bool:
    """Field is an optional type"""
    args = type_.__args__
    return (
        len(args) == 2 and type(None) in args
    )  # pylint: disable=unidiomatic-typecheck


def _resolve_list_from_sub_scripted_type(args: Sequence[Any], options: Options):
    """Handle the various types of sequence type"""
    options.field_args["default"] = list

    (field,) = args
    # Required for Python 3.8
    if field is T:
        options.field_type = ListField

    elif field is not Any and issubclass(field, ResourceBase):
        options.field_args["resource"] = field
        options.field_type = ListOf

    else:
        options.field_args["field"] = process_attribute(field)
        options.field_type = TypedListField


def _resolve_dict_from_sub_scripted_type(args: Sequence[Any], options: Options):
    """Handle the various types of sequence type"""
    options.field_args["default"] = dict

    key_field, value_field = args
    # Required for Python 3.8
    if key_field is KT and value_field is VT:
        options.field_type = DictField

    # Use get origin to determine if value is also sub-scripted. This is
    # required as issubclass cannot be used on sub-scripted fields.
    elif not (get_origin(value_field) or value_field is Any) and issubclass(
        value_field, ResourceBase
    ):
        options.field_args["resource"] = value_field
        options.field_type = DictOf

    else:
        options.field_args["key_field"] = process_attribute(key_field)
        options.field_args["value_field"] = process_attribute(value_field)
        options.field_type = TypedDictField


def _resolve_field_from_sub_scripted_type(origin: Type, options: Options, type_):
    """Resolve a field from a generic type"""
    args = getattr(type_, "__args__", None)
    if not args:
        # This occurs with the plain List and Dict types which are essentially
        # just the native list/dict types
        if origin in (list, dict):
            return _resolve_field_from_annotation(options, origin)

    elif origin is Union:
        # Use nested if as `issubclass` cannot be used with Generic types
        if is_optional(type_):
            options.field_args["null"] = True
            return _resolve_field_from_annotation(options, args[0])
        else:
            pass  # Resolve

    elif origin is Final:
        # Constant
        options.field_type = ConstantField
        value = options.field_args["default"]
        if value is NotProvided:
            raise ResourceDefError(f"Final fields require a value")
        options.base_args["value"] = value
        return

    elif issubclass(origin, List):
        return _resolve_list_from_sub_scripted_type(args, options)

    elif issubclass(origin, Dict):
        return _resolve_dict_from_sub_scripted_type(args, options)

    raise ResourceDefError(f"Unable to resolve field for sub-scripted type {type_}")


def _resolve_field_from_annotation(options: Options, type_):
    """
    Resolve a field from a type annotation

    Handles generics, sequences and optional types
    """
    # Is a sub-scripted type
    if origin := get_origin(type_):
        _resolve_field_from_sub_scripted_type(origin, options, type_)

    # Is a basic type
    elif isinstance(type_, type):
        _resolve_field_from_type(options, type_)

    # Special case
    elif type_ is Any:
        options.field_type = AnyField

    else:
        raise ResourceDefError(f"Annotation is not a type instance {type_}")


def process_attribute(
    type_: Type,
    value: Union[Options, Any] = NotProvided,
    *,
    _base_field: Type[BaseField] = BaseField,
) -> BaseField:
    """Process an individual attribute and generate a field based off values passed"""

    # Attribute is already a field object
    if isinstance(value, _base_field):
        return value

    # Value is a default, wrap with Options
    if not isinstance(value, Options):
        value = Options(value)

    # Resolve field type from annotation
    if not value.field_type:
        _resolve_field_from_annotation(value, type_)

    # Finally instantiate a field object
    return value.init_field()



import pytest

from odin.utils import getmeta
from odin import utils
from .resources import Book


class TestCamelAndDashFormats:
    def test_camel_to_lower_separated(self):
        assert "this_is_an_example" == utils.camel_to_lower_underscore(
            "thisIsAnExample"
        )
        assert "this_is_an_example" == utils.camel_to_lower_underscore(
            "ThisIsAnExample"
        )
        assert "this-is-a-test" == utils.camel_to_lower_dash("thisIsATest")

    def test_lower_underscore_to_camel(self):
        assert "thisIsAnExample" == utils.lower_underscore_to_camel(
            "this_is_an_example"
        )
        assert "thisIsAnExample" == utils.lower_underscore_to_camel(
            "this_is_an_example"
        )
        assert "thisIsATest" == utils.lower_underscore_to_camel("this_Is_A_test")

    def test_lower_dash_to_camel(self):
        assert "thisIsAnExample" == utils.lower_dash_to_camel("this-is-an-example")
        assert "thisIsAnExample" == utils.lower_dash_to_camel("this-is-an-example")
        assert "thisIsATest" == utils.lower_dash_to_camel("this-Is-A-test")


class TestIterator:
    def test_extract_fields_from_dict(self):
        source_dict = {
            "title": "My Title",
            "num_pages": 1,
            "rrp": 1.2,
            "fiction": True,
            "published": ["no_validation_for_type"],
            "another_field": "that is not on the resource",
        }

        assert {
            "title": "My Title",
            "num_pages": 1,
            "rrp": 1.2,
            "fiction": True,
            "published": ["no_validation_for_type"],
        } == utils.extract_fields_from_dict(source_dict, Book)


class TestFilterFields:
    @pytest.fixture
    def field_map(self):
        return getmeta(Book).field_map

    @pytest.mark.parametrize(
        "kwargs, expected_fields, expected_readonly",
        (
            (
                {},
                {
                    "title",
                    "isbn",
                    "num_pages",
                    "rrp",
                    "fiction",
                    "genre",
                    "published",
                    "authors",
                    "publisher",
                },
                set(),
            ),
            ({"include": ["title", "isbn"]}, {"title", "isbn"}, set()),
            (
                {"exclude": ["rrp", "fiction", "genre", "publisher", "authors"]},
                {"title", "isbn", "num_pages", "published"},
                set(),
            ),
            (
                {
                    "include": ["title", "isbn"],
                    "exclude": ["isbn", "rrp", "fiction", "genre"],
                },
                {"title"},
                set(),
            ),
            (
                {"include": ["title", "isbn"], "readonly": ["isbn"]},
                {"title", "isbn"},
                {"isbn"},
            ),
            (
                {"include": ["title", "isbn"], "readonly": ["isbn", "rrp"]},
                {"title", "isbn"},
                {"isbn"},
            ),
            (
                {
                    "include": ["title", "isbn", "rrp"],
                    "exclude": ["rrp"],
                    "readonly": ["isbn", "rrp"],
                },
                {"title", "isbn"},
                {"isbn"},
            ),
        ),
    )
    def test_filtering(self, field_map, kwargs, expected_fields, expected_readonly):
        fields, readonly = utils.filter_fields(field_map, **kwargs)

        assert fields == expected_fields
        assert readonly == expected_readonly


class TestValueInChoice:
    CHOICES = (
        ("foo", "Foo"),
        ("bar", "Foo"),
        (123, "Number"),
        (True, "Boolean"),
        (None, "None"),
        ("eek", "With Help", "Help string"),
    )

    def test_value_in_choices(self):
        assert utils.value_in_choices("foo", self.CHOICES)
        assert utils.value_in_choices(123, self.CHOICES)
        assert utils.value_in_choices(True, self.CHOICES)
        assert utils.value_in_choices(None, self.CHOICES)
        assert utils.value_in_choices("eek", self.CHOICES)
        assert not utils.value_in_choices("xyz", self.CHOICES)
        assert not utils.value_in_choices(False, self.CHOICES)
        assert not utils.value_in_choices(321, self.CHOICES)


class TestForceTuple:
    """
    Tests for odin.utils.force_tuple
    """

    @pytest.mark.parametrize(
        ("value", "expected"),
        (
            (None, ()),
            ("", ("",)),
            (0, (0,)),
            (False, (False,)),
            (True, (True,)),
            (123, (123,)),
            ("Foo", ("Foo",)),
            (["Foo", "bar"], ("Foo", "bar")),
            (("Foo", "bar"), ("Foo", "bar")),
            (["Foo", 123, "bar", "eek"], ("Foo", 123, "bar", "eek")),
        ),
    )
    def test_values(self, value, expected):
        assert utils.force_tuple(value) == expected


class TestChunk:
    """
    Tests for odin.utils.chunk
    """

    @pytest.mark.parametrize(
        ("value", "size", "expected"),
        (
            (
                "This is a block of text split by 4",
                4,
                [
                    ["T", "h", "i", "s"],
                    [" ", "i", "s", " "],
                    ["a", " ", "b", "l"],
                    ["o", "c", "k", " "],
                    ["o", "f", " ", "t"],
                    ["e", "x", "t", " "],
                    ["s", "p", "l", "i"],
                    ["t", " ", "b", "y"],
                    [" ", "4"],
                ],
            ),
        ),
    )
    def test_small_values(self, value, size, expected):
        """Small scale test (eg small iterable that can be easily verified)"""
        assert [list(c) for c in utils.chunk(value, size)] == expected



import datetime

from odin import datetimeutil
from odin import serializers


class TestSerializer:
    def test_date_iso_format(self):
        assert "2014-01-16" == serializers.date_iso_format(datetime.date(2014, 1, 16))

    def test_time_iso_format(self):
        assert "00:29:32+00:00" == serializers.TimeIsoFormat(datetimeutil.utc)(
            datetime.time(0, 29, 32)
        )
        assert "00:29:32+10:00" == serializers.TimeIsoFormat(datetimeutil.utc)(
            datetime.time(
                0, 29, 32, tzinfo=datetimeutil.FixedTimezone.from_hours_minutes(10, 0)
            )
        )
        assert "00:29:32-03:30" == serializers.TimeIsoFormat(
            datetimeutil.FixedTimezone.from_hours_minutes(-3, -30)
        )(datetime.time(0, 29, 32))
        assert "00:29:32+10:00" == serializers.TimeIsoFormat(
            datetimeutil.FixedTimezone.from_hours_minutes(-3, -30)
        )(
            datetime.time(
                0, 29, 32, tzinfo=datetimeutil.FixedTimezone.from_hours_minutes(10, 0)
            )
        )

    def test_datetime_iso_format(self):
        assert "2014-01-16T00:29:32+00:00" == serializers.DatetimeIsoFormat(
            datetimeutil.utc
        )(datetime.datetime(2014, 1, 16, 0, 29, 32))
        assert "2014-01-16T00:29:32+10:00" == serializers.DatetimeIsoFormat(
            datetimeutil.utc
        )(
            datetime.datetime(
                2014,
                1,
                16,
                0,
                29,
                32,
                tzinfo=datetimeutil.FixedTimezone.from_hours_minutes(10, 0),
            )
        )

        assert "2014-01-16T00:29:32-03:30" == serializers.DatetimeIsoFormat(
            datetimeutil.FixedTimezone.from_hours_minutes(-3, -30)
        )(datetime.datetime(2014, 1, 16, 0, 29, 32))
        assert "2014-01-16T00:29:32+10:00" == serializers.DatetimeIsoFormat(
            datetimeutil.FixedTimezone.from_hours_minutes(-3, -30)
        )(
            datetime.datetime(
                2014,
                1,
                16,
                0,
                29,
                32,
                tzinfo=datetimeutil.FixedTimezone.from_hours_minutes(10, 0),
            )
        )



import abc
from typing import Protocol


class ResourceIterable(abc.ABC):
    """
    Iterable object that yields resources.
    """

    @abc.abstractmethod
    def __iter__(self):
        """
        Iterate resources
        """


class TypedResourceIterable(ResourceIterable, abc.ABC):
    """
    Iterable object that yields a specific resource.
    """

    def __init__(self, resource_type):
        self.resource_type = resource_type



import pytest

from odin.resources import create_resource_from_dict
from odin.utils import getmeta
from odin import proxy

from .resources import Book, BookProxy


class MockObject:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


class TestFieldProxyDescriptor:
    @pytest.fixture
    def proxy_obj(self):
        class Target:
            def __init__(self, shadow):
                self._shadow = shadow

        proxy.FieldProxyDescriptor().contribute_to_class(Target, "foo")
        proxy.FieldProxyDescriptor().contribute_to_class(Target, "bar")
        proxy.FieldProxyDescriptor(readonly=True).contribute_to_class(Target, "eek")

        return Target

    def test_get(self, proxy_obj):
        target = proxy_obj(MockObject(foo="123", eek="789"))

        assert target.foo == "123"
        with pytest.raises(AttributeError):
            print(target.bar)

    def test_set(self, proxy_obj):
        shadow = MockObject(foo="123", eek="789")
        target = proxy_obj(shadow)

        target.foo = "321"
        target.bar = "654"

        assert shadow.foo == "321"
        assert shadow.bar == "654"

    def test_set__readonly(self, proxy_obj):
        shadow = MockObject(foo="123", eek="789")
        target = proxy_obj(shadow)

        with pytest.raises(AttributeError):
            target.eek = 321

        assert "789" == shadow.eek


class TestResourceProxyMeta:
    def test_no_meta(self):
        with pytest.raises(AttributeError):

            class SampleProxy(proxy.ResourceProxy):
                pass

    def test_no_resource(self):
        with pytest.raises(AttributeError):

            class SampleProxy(proxy.ResourceProxy):
                class Meta:
                    pass

    def test_repr(self):
        class SampleProxy(proxy.ResourceProxy):
            class Meta:
                resource = Book

        assert "<Proxy of <Options for library.Book>>" == repr(getmeta(SampleProxy))

    def test_field_sorting(self):
        class SampleProxy(proxy.ResourceProxy):
            class Meta:
                field_sorting = True
                resource = Book

        assert [f.attname for f in getmeta(SampleProxy).fields] == [
            "title",
            "isbn",
            "num_pages",
            "rrp",
            "fiction",
            "genre",
            "published",
            "authors",
            "publisher",
        ]

    def test_field_sorting__custom(self):
        class SampleProxy(proxy.ResourceProxy):
            class Meta:
                def field_sorting(fields):
                    return sorted(fields, key=lambda f: f.attname)

                resource = Book

        assert [f.attname for f in getmeta(SampleProxy).fields] == [
            "authors",
            "fiction",
            "genre",
            "isbn",
            "num_pages",
            "published",
            "publisher",
            "rrp",
            "title",
        ]


class TestResourceProxyOptions:
    def test_options_type(self):
        actual = getmeta(BookProxy)
        assert isinstance(actual, proxy.ResourceProxyOptions)

    def test_invalid_option(self):
        with pytest.raises(TypeError):

            class SampleProxy(proxy.ResourceProxy):
                class Meta:
                    resource = Book
                    eek = "boo!"

    @pytest.mark.parametrize(
        "attr, actual",
        (
            ("key_field_names", ("isbn",)),
            ("verbose_name", "Book Summary"),
            ("verbose_name_plural", "Book Summarys"),
            ("include", ("title", "isbn", "num_pages", "rrp")),
            ("name_space", "the.other.library"),
        ),
    )
    def test_options(self, attr, actual):
        target = getmeta(BookProxy)
        assert actual == getattr(target, attr)


class TestResourceProxy:
    def test_shadow(self):
        target = BookProxy()
        actual = target.get_shadow()

        assert isinstance(actual, Book)

    def test_proxy(self):
        actual = Book(title="1984")
        target = BookProxy.proxy(actual)

        assert target.title == "1984"
        assert actual is target.get_shadow()

    def test_proxy__iterable(self):
        actual = [
            Book(title="1984"),
            Book(title="1985 - The man strikes back"),
            Book(title="1986 - Return of the man"),
        ]
        target = BookProxy.proxy(actual)

        assert isinstance(target, proxy.ResourceProxyIter)
        assert ["1984", "1985 - The man strikes back", "1986 - Return of the man"] == [
            t.title for t in target
        ]

    def test_proxy__unknown(self):
        with pytest.raises(TypeError):
            BookProxy.proxy(123)

    def test_extra_property(self):
        book = Book(rrp=256)
        target = BookProxy.proxy(book)

        assert target.expensive

    def test_create_resource_from_dict(self):
        actual = create_resource_from_dict(
            {"title": "1984", "isbn": "1234567", "num_pages": 1}, BookProxy
        )

        assert actual.title == "1984"



import datetime
from odin.codecs import dict_codec
from .resources import *


class TestDictCodec:
    def test_dump(self):
        in_resource = Book(
            title="Consider Phlebas",
            isbn="0-333-45430-8",
            num_pages=471,
            rrp=19.50,
            fiction=True,
            genre="sci-fi",
            authors=[Author(name="Iain M. Banks")],
            publisher=Publisher(name="Macmillan"),
        )

        actual = dict_codec.dump(in_resource)
        assert actual == {
            "$": "library.Book",
            "authors": [
                {
                    "$": "Author",
                    "name": "Iain M. Banks",
                }
            ],
            "fiction": True,
            "genre": "sci-fi",
            "num_pages": 471,
            "rrp": 19.50,
            "isbn": "0-333-45430-8",
            "title": "Consider Phlebas",
            "published": [],
            "publisher": {
                "$": "Publisher",
                "name": "Macmillan",
            },
        }

    def test_dump_no_type_field(self):
        in_resource = Book(
            title="Consider Phlebas",
            isbn="0-333-45430-8",
            num_pages=471,
            rrp=19.50,
            fiction=True,
            genre="sci-fi",
            authors=[Author(name="Iain M. Banks")],
            publisher=Publisher(name="Macmillan"),
        )

        actual = dict_codec.dump(in_resource, include_type_field=False)
        assert actual == {
            "authors": [
                {
                    "name": "Iain M. Banks",
                }
            ],
            "fiction": True,
            "genre": "sci-fi",
            "num_pages": 471,
            "rrp": 19.50,
            "isbn": "0-333-45430-8",
            "title": "Consider Phlebas",
            "published": [],
            "publisher": {
                "name": "Macmillan",
            },
        }

    def test_dump_and_load_(self):
        in_resource = Book(
            title="Consider Phlebas",
            isbn="0-333-45430-8",
            num_pages=471,
            rrp=19.50,
            fiction=True,
            genre="sci-fi",
            authors=[Author(name="Iain M. Banks")],
            publisher=Publisher(name="Macmillan"),
            published=[datetime.datetime(1987, 1, 1, tzinfo=datetime.timezone.utc)],
        )

        d = dict_codec.dump(in_resource)
        out_resource = dict_codec.load(d)

        assert out_resource.title == in_resource.title
        assert out_resource.isbn == in_resource.isbn
        assert out_resource.num_pages == in_resource.num_pages
        assert out_resource.rrp == in_resource.rrp
        assert out_resource.fiction == in_resource.fiction
        assert out_resource.genre == in_resource.genre
        assert out_resource.authors[0].name == in_resource.authors[0].name
        assert out_resource.publisher.name == in_resource.publisher.name
        assert out_resource.published[0] == in_resource.published[0]



import os
from io import StringIO

import pytest
import odin
import odin.exceptions
from odin.codecs import csv_codec

FIXTURE_PATH_ROOT = os.path.join(os.path.dirname(__file__), "fixtures")


class Book(odin.Resource):
    title = odin.StringField(name="Title")
    num_pages = odin.IntegerField(name="Num Pages")
    rrp = odin.FloatField(name="RRP")
    genre = odin.StringField(
        name="Genre",
        choices=(
            ("sci-fi", "Science Fiction"),
            ("fantasy", "Fantasy"),
            ("others", "Others"),
        ),
        null=True,
    )
    author = odin.StringField(name="Author")
    publisher = odin.StringField(name="Publisher")
    language = odin.StringField(name="Language", null=True)

    def extra_attrs(self, attrs):
        self.extras = attrs

    def __eq__(self, other):
        return (
            self.title == other.title
            and self.num_pages == other.num_pages
            and self.rrp == other.rrp
            and (self.genre == other.genre or (not self.genre and not other.genre))
            and self.author == other.author
            and self.publisher == other.publisher
            and (
                self.language == other.language
                or (not self.language and not other.language)
            )
        )


class TestReader:
    def open_fixture(self, file_name):
        return open(os.path.join(FIXTURE_PATH_ROOT, file_name))

    @pytest.mark.parametrize(
        "fixture options".split(),
        (
            ("library-valid.csv", dict(includes_header=True)),
            ("library-header-alt-order.csv", dict(includes_header=True)),
            ("library-last-bad.csv", dict(includes_header=True)),
            ("library-no-header.csv", dict(includes_header=False)),
            (
                "library-lower-header.csv",
                dict(includes_header=True, ignore_header_case=True),
            ),
            ("library-valid.csv", dict(includes_header=True, strict_fields=True)),
        ),
    )
    def test_valid_libraries(self, fixture, options):
        with self.open_fixture(fixture) as f:
            target = csv_codec.reader(f, Book, **options)
            library = list(target)
        assert len(library) == 6
        assert [book.title for book in library] == [
            "Consider Phlebas",
            "The Moonstone",
            "Casino Royale",
            "A Clockwork Orange",
            "The Naked God",
            "Equal Rites",
        ]
        assert [book.num_pages for book in library] == [
            471,
            462,
            181,
            139,
            1256,
            282,
        ]
        assert [book.rrp for book in library] == [
            19.5,
            17.95,
            9.95,
            9.95,
            19.95,
            16.95,
        ]
        genres = [book.genre for book in library]
        assert genres == [
            "sci-fi",
            "others",
            "others",
            "others",
            "sci-fi",
            "fantasy",
        ] or genres == [None, None, None, None, None, None]
        assert [book.author for book in library] == [
            "Iain M. Banks",
            "Wilkie Collins",
            "Ian Fleming",
            "Anthony Burgess",
            "Peter F. Hamilton",
            "Terry Pratchett",
        ]
        assert [book.publisher for book in library] == [
            "Macmillan",
            "Harper & Row",
            "Penguin Books",
            "Penguin Books",
            "Macmillan",
            "Corgi Books",
        ]
        assert [book.language for book in library] == [
            "English",
            "",
            "English",
            "",
            "English",
            "English",
        ]

    def test_extra_fields(self):
        with self.open_fixture("library-header-alt-order.csv") as f:
            target = csv_codec.reader(f, Book, includes_header=True)
            library = list(target)

        assert all(book.extras == ["Other field"] for book in library)

    def test_headers_not_strict(self):
        with self.open_fixture("library-header-alt-order.csv") as f:
            with pytest.raises(odin.exceptions.CodecDecodeError) as result:
                csv_codec.reader(f, Book, includes_header=True, strict_fields=True)

            assert str(result.value) == "Extra unknown fields: Else"

    def test_default_empty_value_is_none(self):
        with self.open_fixture("library-valid.csv") as f:
            target = csv_codec.reader(f, Book, includes_header=True)
            target.default_empty_value = None
            library = list(target)

        assert [book.language for book in library] == [
            "English",
            None,
            "English",
            None,
            "English",
            "English",
        ]

    def test_error(self):
        with self.open_fixture("library-invalid.csv") as f:
            target = csv_codec.reader(f, Book, includes_header=True)

            with pytest.raises(odin.exceptions.ValidationError):
                list(target)

    def test_error_handler(self):
        errors = []

        def error_handler(_, idx):
            errors.append(idx)

        with self.open_fixture("library-invalid.csv") as f:
            target = csv_codec.reader(
                f, Book, includes_header=True, error_callback=error_handler
            )
            library = list(target)

        assert len(library) == 4
        assert len(errors) == 2
        assert errors == [3, 5]

    def test_error_handler_returns_false(self):
        with self.open_fixture("library-invalid.csv") as f:
            target = csv_codec.reader(
                f, Book, includes_header=True, error_callback=lambda v, i: False
            )

            with pytest.raises(odin.exceptions.ValidationError):
                list(target)

    def test_dumps(self):
        with self.open_fixture("library-header-alt-order.csv") as f:
            target = csv_codec.reader(f, Book, includes_header=True)
            expected_library = list(target)

        actual_csv = StringIO(csv_codec.dumps(expected_library, include_header=True))

        actual_library = list(csv_codec.reader(actual_csv, Book, includes_header=True))

        assert sorted(actual_library, key=lambda x: x.num_pages) == sorted(
            expected_library, key=lambda x: x.num_pages
        )

    def test_dump(self, tmpdir):
        with self.open_fixture("library-header-alt-order.csv") as f:
            target = csv_codec.reader(f, Book, includes_header=True)
            expected_library = list(target)

        temp_csv = os.path.join(str(tmpdir), "dump_test.csv")
        with open(temp_csv, "w") as f:
            csv_codec.dump(f, expected_library, include_header=True)

        with open(temp_csv, "r") as f:
            actual_library = list(csv_codec.reader(f, Book, includes_header=True))

        assert sorted(actual_library, key=lambda x: x.num_pages) == sorted(
            expected_library, key=lambda x: x.num_pages
        )



from odin.fields import Field


class AnyField(Field):
    """
    A field that can hold Any (typically that can be handled by the codec) value.

    This might seem counterintuitive however this feature can be used with a typed dict
    to allow for dynamic configuration parameters (although this is discouraged).

    The main use for this field is the Any type for Annotated resources
    """

    data_type_name = "Any"

    def to_python(self, value):
        # Do nothing any value is good
        return value



"""Type aliases for string formatted types."""
from typing import Callable, Union, Any, Sequence, Tuple

__all__ = (
    "Email",
    "IPv4",
    "IPv6",
    "IPv46",
    "Url",
    "Validator",
    "Choices",
)

Validator = Callable[[Any], None]
Choices = Union[
    Sequence[Any],
    Sequence[Tuple[str, Any]],
    Sequence[Tuple[str, Any, str]],
]


class Email(str):
    """Alias for use defining email entries."""

    __slots__ = ()


class IPv4(str):
    """Alias for use defining IPv4 entries."""

    __slots__ = ()


class IPv6(str):
    """Alias for use defining IPv6 entries."""

    __slots__ = ()


class IPv46(str):
    """Alias for use defining IPv4 or IPv6 entries."""

    __slots__ = ()


class Url(str):
    """Alias for use defining URL entries."""

    __slots__ = ()



import pytest

import odin
from . import resources


class TestResourceBase:
    def test_virtual_fields(self):
        target = resources.FromResource()

        assert ["constant_field"] == [f.name for f in target._meta.virtual_fields]

    def test_virtual_field_inheritance(self):
        target = resources.InheritedResource()

        assert ["calculated_field", "constant_field"] == [
            f.name for f in target._meta.virtual_fields
        ]

    def test_fields(self):
        target = resources.FromResource()

        assert [
            "title",
            "count",
            "child",
            "children",
            "excluded1",
            "from_field1",
            "from_field2",
            "from_field3",
            "from_field4",
            "same_but_different",
            "from_field_c1",
            "from_field_c2",
            "from_field_c3",
            "from_field_c4",
            "not_auto_c5",
            "comma_separated_string",
        ] == [f.name for f in target._meta.fields]

    def test_field_inheritance(self):
        target = resources.InheritedResource()

        assert [
            "name",
            "title",
            "count",
            "child",
            "children",
            "excluded1",
            "from_field1",
            "from_field2",
            "from_field3",
            "from_field4",
            "same_but_different",
            "from_field_c1",
            "from_field_c2",
            "from_field_c3",
            "from_field_c4",
            "not_auto_c5",
            "comma_separated_string",
        ] == [f.name for f in target._meta.fields]

    def test_field_multi_inheritance(self):
        target = resources.MultiInheritedResource()

        assert [
            "name",
            "title",
            "count",
            "child",
            "children",
            "excluded1",
            "from_field1",
            "from_field2",
            "from_field3",
            "from_field4",
            "same_but_different",
            "from_field_c1",
            "from_field_c2",
            "from_field_c3",
            "from_field_c4",
            "not_auto_c5",
            "comma_separated_string",
        ] == [f.name for f in target._meta.fields]

    def test_key_field_does_not_exist(self):
        """Ensure an exception is raised if specified key_field is not a member of the resource"""
        with pytest.raises(AttributeError):

            class BadResource(odin.Resource):
                class Meta:
                    key_field_name = "missing_field"

                field_1 = odin.StringField()

    def test_key_fields_from_field_attributes(self):
        """
        Ensure that multiple keys can be defined from fields and in the right order.
        """
        assert ["name", "genre"] == [
            f.attname for f in resources.OldBook._meta.key_fields
        ]



import datetime
from typing import Union, Callable, Any, Dict

from odin import datetimeutil
from odin.compatibility import deprecated

StringFormatter = Callable[[Any], str]
STRING_TYPES: Dict[type, StringFormatter] = {}


def register_string_type(data_type: type, serialise_method: StringFormatter):
    """
    Register a data type serialise method that results in a string.

    :param data_type: a type instance that string data type should match.
    :param serialise_method: a method that accepts a single value that will return a string representation of
        the serialised type.

    """
    STRING_TYPES[data_type] = serialise_method


def date_iso_format(value: datetime.date) -> str:
    """
    Serialise a datetime.date to ISO string format.
    """
    return value.isoformat()


@deprecated(
    "Defaulting the timezone should be preformed by the fields only, this confuses things. "
    "Most codecs have already migrated to not use this class."
)
class DatetimeIsoFormat:
    """
    Serialise a datetime.time or datetime.datetime to ISO string format.
    """

    def __init__(self, default_timezone=datetimeutil.local):
        self.default_timezone = default_timezone

    def __call__(self, value: Union[datetime.time, datetime.datetime]) -> str:
        if value.tzinfo is None:
            value = value.replace(tzinfo=self.default_timezone)
        return value.isoformat()


TimeIsoFormat = DatetimeIsoFormat

# Mapping for compatibility
datetime_iso_format = datetime.datetime.isoformat
time_iso_format = datetime.time.isoformat



import datetime
import json

from odin.codecs import json_codec
from odin.utils import getmeta

from tests.resources_annotated import Book, Publisher, Author, Library, IdentifiableBook


class TestAnnotatedResourceType:
    def test_fields_are_identified_in_correct_order(self):
        meta = getmeta(Book)

        assert [f.name for f in meta.fields] == [
            "title",
            "isbn",
            "num_pages",
            "rrp",
            "fiction",
            "genre",
            "published",
            "authors",
            "publisher",
        ]


class TestAnnotatedResource:
    def test_fields_are_inherited_from_parent_resources(self):
        meta = getmeta(IdentifiableBook)

        assert [f.name for f in meta.fields] == [
            "title",
            "isbn",
            "num_pages",
            "rrp",
            "fiction",
            "genre",
            "published",
            "authors",
            "publisher",
            "id",
            "purchased_from",
        ]


class TestAnnotatedKitchenSink:
    def test_dumps_with_valid_data(self):
        book = Book(
            title="Consider Phlebas",
            isbn="0-333-45430-8",
            num_pages=471,
            rrp=19.50,
            genre="sci-fi",
            fiction=True,
            published=[datetime.datetime(1987, 1, 1, tzinfo=datetime.timezone.utc)],
        )
        book.publisher = Publisher(name="Macmillan")
        book.authors.append(Author(name="Iain M. Banks"))

        library = Library(name="Public Library", books=[book])
        actual = json_codec.dumps(library)

        assert json.loads(actual) == json.loads(
            """
{
    "$": "new_library.Library",
    "name": "Public Library",
    "books": [
        {
            "$": "new_library.Book",
            "publisher": {
                "$": "Publisher",
                "name": "Macmillan"
            },
            "num_pages": 471,
            "isbn": "0-333-45430-8",
            "title": "Consider Phlebas",
            "authors": [
                {
                    "$": "Author",
                    "name": "Iain M. Banks"
                }
            ],
            "fiction": true,
            "published": ["1987-01-01T00:00:00+00:00"],
            "genre": "sci-fi",
            "rrp": 19.5
        }
    ],
    "subscribers": null,
    "book_count": 1
}
        """
        )



import pytest

import odin
from odin import traversal
from odin.exceptions import NoMatchError, InvalidPathError, MultipleMatchesError


class Level3(odin.Resource):
    class Meta:
        key_field_name = "name"
        namespace = "odin.traversal"

    name = odin.StringField()


class Level2(odin.Resource):
    class Meta:
        namespace = "odin.traversal"

    name = odin.StringField()
    label = odin.StringField(null=True)
    level3s = odin.ListOf(Level3)


class Level1(odin.Resource):
    class Meta:
        namespace = "odin.traversal"

    name = odin.StringField()
    level2 = odin.DictAs(Level2)
    level2s = odin.DictOf(Level2)


TEST_STRUCTURE = Level1(
    name="a",
    level2=Level2(name="b", level3s=[]),
    level2s=dict(
        a=Level2(name="c", level3s=[]),
        b=Level2(
            name="d", label="not_empty", level3s=[Level3(name="e"), Level3(name="f")]
        ),
        c=Level2(name="g", label="not_empty", level3s=[Level3(name="h")]),
    ),
)

TEST_LIST_STRUCTURE = [
    Level1(
        name="a",
        level2=Level2(name="b", level3s=[]),
        level2s=dict(
            a=Level2(name="c", level3s=[]),
            b=Level2(name="d", level3s=[Level3(name="e"), Level3(name="f")]),
            c=Level2(name="g", level3s=[Level3(name="h")]),
        ),
    ),
    Level1(
        name="i",
        level2=Level2(name="j", level3s=[]),
        level2s=dict(
            a=Level2(name="k", level3s=[]),
            b=Level2(name="l", level3s=[Level3(name="m"), Level3(name="n")]),
            c=Level2(name="o", level3s=[Level3(name="p")]),
        ),
    ),
]


class ResourceTraversalIteratorTest(traversal.ResourceTraversalIterator):
    def __init__(self, resource):
        super(ResourceTraversalIteratorTest, self).__init__(resource)
        self.events = []

    def on_pre_enter(self):
        self.events.append("on_pre_enter: %s" % self.path)

    def on_enter(self):
        self.events.append("on_enter: %s" % self.path)

    def on_exit(self):
        self.events.append("on_exit: %s" % self.path)


class TestTraversal:
    def test_structure(self):
        TEST_STRUCTURE.full_clean()

        resource_iter = ResourceTraversalIteratorTest(TEST_STRUCTURE)
        resources = [
            "%s %s %s" % (r, r.name, resource_iter.depth) for r in resource_iter
        ]

        assert [
            "on_enter: ",
            "on_enter: level2",
            "on_exit: level2",
            "on_enter: level2s[a]",
            "on_exit: level2s[a]",
            "on_enter: level2s[b]",
            "on_enter: level2s[b].level3s{name=e}",
            "on_exit: level2s[b].level3s{name=e}",
            "on_enter: level2s[b].level3s{name=f}",
            "on_exit: level2s[b].level3s{name=f}",
            "on_exit: level2s[b]",
            "on_enter: level2s[c]",
            "on_enter: level2s[c].level3s{name=h}",
            "on_exit: level2s[c].level3s{name=h}",
            "on_exit: level2s[c]",
            "on_exit: ",
        ] == resource_iter.events

        assert [
            "odin.traversal.Level1 resource a 0",
            "odin.traversal.Level2 resource b 1",
            "odin.traversal.Level2 resource c 1",
            "odin.traversal.Level2 resource d 1",
            "odin.traversal.Level3 resource e 2",
            "odin.traversal.Level3 resource f 2",
            "odin.traversal.Level2 resource g 1",
            "odin.traversal.Level3 resource h 2",
        ] == resources

    def test_list_structure(self):
        TEST_STRUCTURE.full_clean()

        resource_iter = ResourceTraversalIteratorTest(TEST_LIST_STRUCTURE)
        resources = [
            "%s %s %s" % (r, r.name, resource_iter.depth) for r in resource_iter
        ]

        assert [
            "on_enter: ",
            "on_enter: level2",
            "on_exit: level2",
            "on_enter: level2s[a]",
            "on_exit: level2s[a]",
            "on_enter: level2s[b]",
            "on_enter: level2s[b].level3s{name=e}",
            "on_exit: level2s[b].level3s{name=e}",
            "on_enter: level2s[b].level3s{name=f}",
            "on_exit: level2s[b].level3s{name=f}",
            "on_exit: level2s[b]",
            "on_enter: level2s[c]",
            "on_enter: level2s[c].level3s{name=h}",
            "on_exit: level2s[c].level3s{name=h}",
            "on_exit: level2s[c]",
            "on_exit: ",
            "on_enter: ",
            "on_enter: level2",
            "on_exit: level2",
            "on_enter: level2s[a]",
            "on_exit: level2s[a]",
            "on_enter: level2s[b]",
            "on_enter: level2s[b].level3s{name=m}",
            "on_exit: level2s[b].level3s{name=m}",
            "on_enter: level2s[b].level3s{name=n}",
            "on_exit: level2s[b].level3s{name=n}",
            "on_exit: level2s[b]",
            "on_enter: level2s[c]",
            "on_enter: level2s[c].level3s{name=p}",
            "on_exit: level2s[c].level3s{name=p}",
            "on_exit: level2s[c]",
            "on_exit: ",
        ] == resource_iter.events

        assert [
            "odin.traversal.Level1 resource a 0",
            "odin.traversal.Level2 resource b 1",
            "odin.traversal.Level2 resource c 1",
            "odin.traversal.Level2 resource d 1",
            "odin.traversal.Level3 resource e 2",
            "odin.traversal.Level3 resource f 2",
            "odin.traversal.Level2 resource g 1",
            "odin.traversal.Level3 resource h 2",
            "odin.traversal.Level1 resource i 0",
            "odin.traversal.Level2 resource j 1",
            "odin.traversal.Level2 resource k 1",
            "odin.traversal.Level2 resource l 1",
            "odin.traversal.Level3 resource m 2",
            "odin.traversal.Level3 resource n 2",
            "odin.traversal.Level2 resource o 1",
            "odin.traversal.Level3 resource p 2",
        ] == resources


class TestTraversalPath:
    def test_parse(self):
        actual = traversal.TraversalPath.parse("level2")
        assert (
            traversal.TraversalPath(
                (traversal.NotSupplied, traversal.NotSupplied, "level2"),
            )
            == actual
        )

        actual = traversal.TraversalPath.parse("level2.name")
        assert (
            traversal.TraversalPath(
                (traversal.NotSupplied, traversal.NotSupplied, "level2"),
                (traversal.NotSupplied, traversal.NotSupplied, "name"),
            )
            == actual
        )

        actual = traversal.TraversalPath.parse("level2s[b].level3s[1].name")
        assert (
            traversal.TraversalPath(
                ("b", traversal.NotSupplied, "level2s"),
                ("1", traversal.NotSupplied, "level3s"),
                (traversal.NotSupplied, traversal.NotSupplied, "name"),
            )
            == actual
        )

        actual = traversal.TraversalPath.parse("level2s[b].level3s{code=abc}.name")
        assert (
            traversal.TraversalPath(
                ("b", traversal.NotSupplied, "level2s"),
                ("abc", "code", "level3s"),
                (traversal.NotSupplied, traversal.NotSupplied, "name"),
            )
            == actual
        )

    def test_add(self):
        actual = traversal.TraversalPath.parse("level2") + "name"
        assert traversal.TraversalPath.parse("level2.name") == actual

        actual = traversal.TraversalPath.parse(
            "level2s[b]"
        ) + traversal.TraversalPath.parse("level3s[1].name")
        assert traversal.TraversalPath.parse("level2s[b].level3s[1].name") == actual

    def test_valid_path(self):
        assert "a" == traversal.TraversalPath.parse("name").get_value(TEST_STRUCTURE)
        assert "b" == traversal.TraversalPath.parse("level2.name").get_value(
            TEST_STRUCTURE
        )

        r = traversal.TraversalPath.parse("level2s[b].level3s[1]").get_value(
            TEST_STRUCTURE
        )
        assert isinstance(r, Level3)
        assert "f" == r.name

        r = traversal.TraversalPath.parse("level2s[b].level3s{name=f}").get_value(
            TEST_STRUCTURE
        )
        assert isinstance(r, Level3)
        assert "f" == r.name

        r = traversal.TraversalPath.parse("level2s{name=g}.level3s{name=h}").get_value(
            TEST_STRUCTURE
        )
        assert isinstance(r, Level3)
        assert "h" == r.name

    def test_invalid_path(self):
        path = traversal.TraversalPath.parse("level2s[b].level3s[4]")
        pytest.raises(NoMatchError, path.get_value, TEST_STRUCTURE)

        path = traversal.TraversalPath.parse("level2s[b].level3s_sd[1]")
        pytest.raises(InvalidPathError, path.get_value, TEST_STRUCTURE)

        path = traversal.TraversalPath.parse("level2s[d].level3s{name=h}")
        pytest.raises(NoMatchError, path.get_value, TEST_STRUCTURE)

        path = traversal.TraversalPath.parse("level2s{label=not_empty}.level3s{name=h}")
        pytest.raises(MultipleMatchesError, path.get_value, TEST_STRUCTURE)



from odin import adapters
from odin.utils import field_iter_items
from .resources import *


class TestResourceOptionsAdapter:
    def test_repr(self):
        options = adapters.ResourceAdapter(Book())._meta

        assert "<Options Adapter for library.Book>" == repr(options)

    def test_attribute_fields(self):
        options = adapters.ResourceAdapter(Book())._meta

        assert ["fiction"] == [f.name for f in options.attribute_fields]

    def test_elements_fields(self):
        options = adapters.ResourceAdapter(Book())._meta

        assert [
            "title",
            "isbn",
            "num_pages",
            "rrp",
            "genre",
            "published",
            "authors",
            "publisher",
        ] == [f.name for f in options.element_fields]

    def test_attribute_proxying(self):
        options = adapters.ResourceAdapter(Book())._meta

        assert "Book" == options.verbose_name


class TestResourceAdapter:
    def test_field_proxy(self):
        target = adapters.ResourceAdapter(Book(title="Foo"))

        assert "Foo" == target.title

    def test_include_fields(self):
        book = Book(title="Foo", rrp=123.45, num_pages=10, fiction=True)

        target = adapters.ResourceAdapter(book, include=("title", "rrp"))

        assert ["rrp", "title"] == sorted(target._meta.field_map.keys())

    def test_exclude_fields(self):
        book = Book(title="Foo", rrp=123.45, num_pages=10, fiction=True)

        target = adapters.ResourceAdapter(book, exclude=("title", "rrp"))

        assert [
            "authors",
            "fiction",
            "genre",
            "isbn",
            "num_pages",
            "published",
            "publisher",
        ] == sorted(target._meta.field_map.keys())

    def test_repr(self):
        target = adapters.ResourceAdapter(Book(title="Foo"))

        assert "<ResourceAdapter: library.Book resource adapter>" == repr(target)

    def test_iter(self):
        book = Book(title="Foo", isbn="abc123", rrp=123.45, num_pages=10, fiction=True)
        target = adapters.ResourceAdapter(book)

        actual = list((f.name, str(v)) for f, v in field_iter_items(target))
        assert [
            ("title", "Foo"),
            ("isbn", "abc123"),
            ("num_pages", "10"),
            ("rrp", "123.45"),
            ("fiction", "True"),
            ("genre", "None"),
            ("published", "[]"),
            ("authors", "[]"),
            ("publisher", "None"),
        ] == actual

    def test_to_dict(self):
        publisher = Publisher()
        book = Book(
            title="Foo",
            isbn="abc123",
            rrp=123.45,
            num_pages=10,
            fiction=True,
            publisher=publisher,
        )
        target = adapters.ResourceAdapter(book)

        assert {
            "authors": [],
            "fiction": True,
            "genre": None,
            "num_pages": 10,
            "published": [],
            "publisher": publisher,
            "rrp": 123.45,
            "isbn": "abc123",
            "title": "Foo",
        } == target.to_dict()

    def test_apply_to(self):
        sources = [Library(name="Foo"), Publisher(name="Bar"), Library(name="Eek")]
        result = adapters.ResourceAdapter.apply_to(sources, include=["name"])
        actuals = list(result)

        assert [{"name": "Foo"}, {"name": "Bar"}, {"name": "Eek"}] == [
            a.to_dict() for a in actuals
        ]

    def test_curried(self):
        book = Book(
            title="Foo",
            isbn="abc123",
            rrp=123.45,
            num_pages=10,
            fiction=True,
            publisher=Publisher(),
        )
        SummaryAdapter = adapters.ResourceAdapter.curry(
            include=("title", "rrp", "fiction")
        )
        target = SummaryAdapter(book)

        assert {"fiction": True, "rrp": 123.45, "title": "Foo"} == target.to_dict()



import pytest

import odin
from odin.exceptions import ValidationError
from odin.fields.composite import DictAs, ArrayOf, DictOf


class ExampleResource(odin.Resource):
    name = odin.StringField()


class TestFields:
    def assertResourceEqual(self, first, second):
        if not first:
            raise AssertionError("First item is None")
        if not second:
            raise AssertionError("Second item is None")

        if first._meta.resource_name != second._meta.resource_name:
            raise AssertionError(
                "Resources are different types %s != %s"
                % (first._meta.resource_name, second._meta.resource_name)
            )

        for f in first._meta.fields:
            first_val = getattr(first, f.name)
            second_val = getattr(second, f.name)
            if first_val != second_val:
                raise AssertionError(
                    "Resource field %s does not match '%s' != '%s'"
                    % (f.name, first_val, second_val)
                )

    def assertResourceListEqual(self, first, second):
        if not isinstance(first, list):
            raise AssertionError("First is not a list")
        if not isinstance(second, list):
            raise AssertionError("Second is not a list")
        if len(first) != len(second):
            raise AssertionError("Lists are different lengths")
        for idx in range(len(first)):
            self.assertResourceEqual(first[idx], second[idx])

    def assertResourceDictEqual(self, first, second):
        if not isinstance(first, dict):
            raise AssertionError("First is not a dict")
        if not isinstance(second, dict):
            raise AssertionError("Second is dict")
        if len(first) != len(second):
            raise AssertionError("Dicts are different lengths")
        for key in first.keys():
            self.assertResourceEqual(first[key], second[key])

    # DictAs ##################################################################

    def test_dictas_ensure_is_resource(self):
        with pytest.raises(TypeError):
            DictAs("an item")

    def test_dictas_1(self):
        f = DictAs(ExampleResource)
        pytest.raises(ValidationError, f.clean, None)
        pytest.raises(ValidationError, f.clean, "abc")
        pytest.raises(ValidationError, f.clean, 123)
        pytest.raises(ValidationError, f.clean, {})
        pytest.raises(
            ValidationError,
            f.clean,
            {"$": "tests.test_fields_composite.ExampleResource"},
        )
        self.assertResourceEqual(ExampleResource(name="foo"), f.clean({"name": "foo"}))
        self.assertResourceEqual(
            ExampleResource(name="foo"),
            f.clean(
                {"$": "tests.test_fields_composite.ExampleResource", "name": "foo"}
            ),
        )

    def test_dictas_2(self):
        f = DictAs(ExampleResource, null=True)
        assert f.clean(None) is None
        pytest.raises(ValidationError, f.clean, "abc")
        pytest.raises(ValidationError, f.clean, 123)
        pytest.raises(ValidationError, f.clean, {})
        pytest.raises(
            ValidationError,
            f.clean,
            {"$": "tests.test_fields_composite.ExampleResource"},
        )
        self.assertResourceEqual(ExampleResource(name="foo"), f.clean({"name": "foo"}))
        self.assertResourceEqual(
            ExampleResource(name="foo"),
            f.clean(
                {"$": "tests.test_fields_composite.ExampleResource", "name": "foo"}
            ),
        )

    # ArrayOf #################################################################

    def test_arrayof_1(self):
        f = ArrayOf(ExampleResource)
        pytest.raises(ValidationError, f.clean, None)
        pytest.raises(ValidationError, f.clean, "abc")
        pytest.raises(ValidationError, f.clean, 123)
        pytest.raises(ValidationError, f.clean, {})
        pytest.raises(
            ValidationError,
            f.clean,
            {"$": "tests.test_fields_composite.ExampleResource"},
        )
        pytest.raises(ValidationError, f.clean, [None])
        self.assertResourceListEqual([], f.clean([]))
        self.assertResourceListEqual(
            [ExampleResource(name="foo")], f.clean([{"name": "foo"}])
        )
        self.assertResourceListEqual(
            [ExampleResource(name="foo")],
            f.clean(
                [{"$": "tests.test_fields_composite.ExampleResource", "name": "foo"}]
            ),
        )

    def test_arrayof_2(self):
        f = ArrayOf(ExampleResource, null=True)
        assert f.clean(None) is None
        pytest.raises(ValidationError, f.clean, "abc")
        pytest.raises(ValidationError, f.clean, 123)
        pytest.raises(ValidationError, f.clean, {})
        pytest.raises(
            ValidationError,
            f.clean,
            {"$": "tests.test_fields_composite.ExampleResource"},
        )
        pytest.raises(ValidationError, f.clean, [None])
        self.assertResourceListEqual([], f.clean([]))
        self.assertResourceListEqual(
            [ExampleResource(name="foo")], f.clean([{"name": "foo"}])
        )
        self.assertResourceListEqual(
            [ExampleResource(name="foo")],
            f.clean(
                [{"$": "tests.test_fields_composite.ExampleResource", "name": "foo"}]
            ),
        )

    def test_arrayof_3(self):
        f = ArrayOf(ExampleResource, empty=False)
        pytest.raises(ValidationError, f.clean, None)
        pytest.raises(ValidationError, f.clean, "abc")
        pytest.raises(ValidationError, f.clean, 123)
        pytest.raises(ValidationError, f.clean, {})
        pytest.raises(
            ValidationError,
            f.clean,
            {"$": "tests.test_fields_composite.ExampleResource"},
        )
        pytest.raises(ValidationError, f.clean, [None])
        pytest.raises(ValidationError, f.clean, [])
        self.assertResourceListEqual(
            [ExampleResource(name="foo")], f.clean([{"name": "foo"}])
        )
        self.assertResourceListEqual(
            [ExampleResource(name="foo")],
            f.clean(
                [{"$": "tests.test_fields_composite.ExampleResource", "name": "foo"}]
            ),
        )

    # DictOf #################################################################

    def test_dictof_1(self):
        f = DictOf(ExampleResource)
        pytest.raises(ValidationError, f.clean, None)
        pytest.raises(ValidationError, f.clean, "abc")
        pytest.raises(ValidationError, f.clean, 123)
        pytest.raises(ValidationError, f.clean, [])
        pytest.raises(
            ValidationError,
            f.clean,
            {"$": "tests.test_fields_composite.ExampleResource"},
        )
        pytest.raises(ValidationError, f.clean, {"abc": None})
        self.assertResourceDictEqual({}, f.clean({}))
        self.assertResourceDictEqual(
            {"foo": ExampleResource(name="foo")}, f.clean({"foo": {"name": "foo"}})
        )
        self.assertResourceDictEqual(
            {"foo": ExampleResource(name="foo")},
            f.clean(
                {
                    "foo": {
                        "$": "tests.test_fields_composite.ExampleResource",
                        "name": "foo",
                    }
                }
            ),
        )

    def test_dictof_2(self):
        f = DictOf(ExampleResource, null=True)
        assert f.clean(None) is None
        pytest.raises(ValidationError, f.clean, "abc")
        pytest.raises(ValidationError, f.clean, 123)
        pytest.raises(ValidationError, f.clean, [])
        pytest.raises(
            ValidationError,
            f.clean,
            {"$": "tests.test_fields_composite.ExampleResource"},
        )
        pytest.raises(ValidationError, f.clean, {"abc": None})
        self.assertResourceDictEqual({}, f.clean({}))
        self.assertResourceDictEqual(
            {"foo": ExampleResource(name="foo")}, f.clean({"foo": {"name": "foo"}})
        )
        self.assertResourceDictEqual(
            {"foo": ExampleResource(name="foo")},
            f.clean(
                {
                    "foo": {
                        "$": "tests.test_fields_composite.ExampleResource",
                        "name": "foo",
                    }
                }
            ),
        )

    def test_dictof_3(self):
        f = DictOf(ExampleResource, empty=False)
        pytest.raises(ValidationError, f.clean, None)
        pytest.raises(ValidationError, f.clean, "abc")
        pytest.raises(ValidationError, f.clean, 123)
        pytest.raises(ValidationError, f.clean, [])
        pytest.raises(
            ValidationError,
            f.clean,
            {"$": "tests.test_fields_composite.ExampleResource"},
        )
        pytest.raises(ValidationError, f.clean, {"abc": None})
        pytest.raises(ValidationError, f.clean, {})
        self.assertResourceDictEqual(
            {"foo": ExampleResource(name="foo")}, f.clean({"foo": {"name": "foo"}})
        )
        self.assertResourceDictEqual(
            {"foo": ExampleResource(name="foo")},
            f.clean(
                {
                    "foo": {
                        "$": "tests.test_fields_composite.ExampleResource",
                        "name": "foo",
                    }
                }
            ),
        )

    def test_dictof_key_choices(self):
        f = DictOf(
            ExampleResource,
            null=True,
            key_choices=(
                ("foo", "Foo"),
                (None, "None"),
            ),
        )
        pytest.raises(ValidationError, f.clean, {"eek": ExampleResource(name="foo")})
        self.assertResourceDictEqual(
            {"foo": ExampleResource(name="foo")}, f.clean({"foo": {"name": "foo"}})
        )



import pytest

import odin
from odin.fields import NotProvided
from odin.resources import (
    ResourceOptions,
    build_object_graph,
    create_resource_from_dict,
    Resource,
)
from odin.exceptions import ValidationError, ResourceDefError
from odin.utils import getmeta
from .resources import Book, BookProxy, Library, Subscriber


class Author(odin.Resource):
    name = odin.StringField()
    country = odin.StringField(null=True)

    def clean(self):
        if self.name == "Bruce" and self.country.startswith("Australia"):
            raise ValidationError("No no no no")

    def clean_country(self, value):
        if value not in ["Australia", "New Zealand"]:
            raise ValidationError("What are ya?")
        return "%s!" % value

    def extra_attrs(self, attrs):
        self.extras = attrs


# Resources for abstract namespace test.
class ResourceA(odin.Resource):
    class Meta:
        abstract = True
        namespace = "example"

    a = odin.StringField()


class ResourceB(ResourceA):
    class Meta:
        abstract = True

    b = odin.StringField()


class ResourceC(ResourceB):
    c = odin.StringField()


class ResourceD(ResourceC):
    class Meta:
        field_sorting = True

    d = odin.StringField()


class ResourceE(ResourceD):
    class Meta:
        def field_sorting(fields):
            return sorted(fields, key=lambda f: f.name)

    aa = odin.StringField()


class ResourceF(ResourceE):
    aaa = odin.StringField()


class ResourceG(odin.Resource):
    """
    Test with user metadata
    """

    class Meta:
        user_data = {"custom": "my-custom-metadata"}

    name = odin.StringField()


class TestResource:
    def test_constructor_kwargs_only(self):
        r = Author(name="Foo")
        assert "Foo" == r.name

    def test_constructor_kwargs_only_with_unknown_field(self):
        with pytest.raises(TypeError):
            Author(name="Foo", age=42)

    def test_constructor_args_only(self):
        r = Author("Foo")
        assert "Foo" == r.name
        assert None == r.country

    def test_constructor_args_excess(self):
        with pytest.raises(TypeError) as ex:
            Author("Foo", "Australia", 42)
        assert "This resource takes 2 positional arguments but 3 where given." == str(
            ex.value
        )

    def test_constructor_args_and_kwargs(self):
        r = Author("Foo", country="Australia")
        assert "Foo" == r.name
        assert "Australia" == r.country

    def test_constructor_args_and_kwargs_overrides(self):
        r = Author("Foo", name="Bar", country="Australia")
        assert "Foo" == r.name
        assert "Australia" == r.country

    def test_simple_methods(self):
        r = Author()
        assert "<Author: tests.test_resources.Author resource>" == repr(r)
        assert "tests.test_resources.Author resource" == str(r)

    def test_clean_fields_1(self):
        r = Author(name="Foo")
        r.country = "Australia"

        r.clean_fields()

        assert "Foo" == r.name
        assert "Australia!" == r.country

    def test_clean_fields_2(self):
        r = Author(name="Foo")

        r.clean_fields()

        assert "Foo" == r.name
        assert None == r.country

    def test_clean_fields_3(self):
        r = Author(name="Foo", country="England")

        try:
            r.clean_fields()
        except ValidationError as ve:
            assert ["What are ya?"] == ve.message_dict["country"]
        else:
            raise AssertionError("ValidationError not raised.")

    def test_clean_fields_4(self):
        r = Author()

        try:
            r.clean_fields()
        except ValidationError as ve:
            assert ["This field cannot be null."] == ve.message_dict["name"]
        else:
            raise AssertionError("ValidationError not raised.")

    def test_full_clean(self):
        r = Author(name="Bruce", country="Australia")

        try:
            r.full_clean()
        except ValidationError as ve:
            assert ["No no no no"] == ve.message_dict["__all__"]
        else:
            raise AssertionError("ValidationError not raised.")

    def test_full_clean_exclude(self):
        r = Author(name="Bruce", country="England")

        r.full_clean(exclude=("country",))

        assert "Bruce" == r.name
        assert "England" == r.country

    # Fix for #11
    def test_multiple_abstract_namespaces(self):
        assert "example.ResourceC" == ResourceC._meta.resource_name

    def test_parents_1(self):
        assert [] == ResourceA._meta.parents
        assert [ResourceA] == ResourceB._meta.parents
        assert [ResourceA, ResourceB] == ResourceC._meta.parents
        assert [ResourceA, ResourceB, ResourceC] == ResourceD._meta.parents

    def test_field_sorting(self):
        assert ["c", "b", "a"] == [f.name for f in ResourceC._meta.fields]

    def test_field_sorting__enabled(self):
        assert ["a", "b", "c", "d"] == [f.name for f in ResourceD._meta.fields]

    def test_field_sorting__callable(self):
        assert ["a", "aa", "b", "c", "d"] == [f.name for f in ResourceE._meta.fields]

    def test_field_sorting__inherited(self):
        assert ["a", "aa", "aaa", "b", "c", "d"] == [
            f.name for f in ResourceF._meta.fields
        ]

    def test_user_data__where_not_supplied(self):
        assert getmeta(ResourceA).user_data is None

    def test_user_data__where_supplied_on_type(self):
        assert getmeta(ResourceG).user_data == {"custom": "my-custom-metadata"}

    def test_user_data__where_supplied_on_instance(self):
        target = ResourceG(name="Ludwig")
        assert getmeta(target).user_data == {"custom": "my-custom-metadata"}


class TestMetaOptions:
    def test_invalid_options(self):
        class Meta:
            random_val = 10

        class NewResource:
            pass

        target = ResourceOptions(Meta)

        with pytest.raises(TypeError):
            target.contribute_to_class(NewResource, "etc")

    def test_use_a_reserved_field(self):
        with pytest.raises(ResourceDefError, match="fields"):

            class InvalidFieldsResource(Resource):
                fields = odin.StringField()


class TestConstructionMethods:
    def test_build_object_graph_empty_dict_no_clean(self):
        book = build_object_graph({}, Book, full_clean=False)

        assert (
            dict(
                title=None,
                isbn=None,
                num_pages=None,
                rrp=20.4,
                fiction=None,
                genre=None,
                published=None,
                authors=None,
                publisher=None,
            )
            == book.to_dict()
        )

    def test_build_object_graph_empty_dict(self):
        with pytest.raises(ValidationError) as ctx:
            build_object_graph({}, Book)

        assert (
            dict(
                title=["This field cannot be null."],
                isbn=["This field cannot be null."],
                num_pages=["This field cannot be null."],
                fiction=["This field cannot be null."],
                genre=["This field cannot be null."],
                published=["This field cannot be null."],
                authors=["List cannot contain null entries."],
            )
            == ctx.value.error_messages
        )

    def test_build_object_graph_from_list(self):
        books = build_object_graph(
            [dict(title="Book1"), dict(title="Book2")], Book, full_clean=False
        )

        assert [
            dict(
                title="Book1",
                isbn=None,
                num_pages=None,
                rrp=20.4,
                fiction=None,
                genre=None,
                published=None,
                authors=None,
                publisher=None,
            ),
            dict(
                title="Book2",
                isbn=None,
                num_pages=None,
                rrp=20.4,
                fiction=None,
                genre=None,
                published=None,
                authors=None,
                publisher=None,
            ),
        ] == [book.to_dict() for book in books]

    def test_build_nested_objects(self):
        subscribers = [
            {"name": "John Smith", "address": "Oak Lane 1234"},
            {"name": "Johnny Smith", "address": "Oak Lane 1235"},
        ]

        library = {"name": "John Smith Library", "subscribers": subscribers}

        expected = sorted(
            build_object_graph(subscribers, resource=Subscriber, full_clean=False),
            key=lambda s: s.name,
        )
        actual = sorted(
            build_object_graph(library, resource=Library, full_clean=False).subscribers,
            key=lambda s: s.name,
        )

        assert actual == expected

    def test_build_nested_objects_with_polymorphism(self):
        books = [
            {
                "title": "Book1",
                "isbn": "abc-123",
                "num_pages": 1,
                "rrp": 20.4,
                "fiction": True,
                "genre": "sci-fi",
                "published": [],
                "authors": [],
                "publisher": None,
                "$": "tests.resources.Book",
            },
            {
                "title": "Book2",
                "isbn": "def-456",
                "num_pages": 1,
                "rrp": 20.4,
                "fiction": True,
                "genre": "sci-fi",
                "published": [],
                "authors": [],
                "publisher": None,
                "$": "tests.resources.Book",
            },
        ]

        library = {"name": "John Smith Library", "books": books}

        expected = sorted(
            build_object_graph(books, full_clean=False), key=lambda s: s.title
        )
        actual = sorted(
            build_object_graph(library, resource=Library, full_clean=False).books,
            key=lambda s: s.title,
        )

        assert actual == expected

    def test_create_resource_from_dict_with_default_to_not_supplied(self):
        book = create_resource_from_dict(
            {"title": "Foo", "num_pages": 42},
            Book,
            full_clean=False,
            default_to_not_provided=True,
        )

        assert (
            dict(
                title="Foo",
                isbn=NotProvided,
                num_pages=42,
                rrp=NotProvided,
                fiction=NotProvided,
                genre=NotProvided,
                published=NotProvided,
                authors=NotProvided,
                publisher=NotProvided,
            )
            == book.to_dict()
        )

    def test_create_resource_from_dict__with_proxy(self):
        book = create_resource_from_dict(
            {"title": "Foo", "num_pages": 42, "rrp": 10000.99, "fiction": True},
            BookProxy,
            full_clean=False,
            default_to_not_provided=True,
        )

        assert (
            dict(
                title="Foo",
                isbn=NotProvided,
                num_pages=42,
                rrp=20.4,
            )
            == book.to_dict()
        )



"""
Tests of polymorphic behaviours
"""
import pytest

import odin
from odin.resources import create_resource_from_dict
from odin.utils import getmeta


class AbstractResource(odin.Resource):
    class Meta:
        abstract = True
        namespace = "au.com.example.abstracts"
        type_field = "type"


class ResourceA(AbstractResource):
    class Meta:
        type_field = "type"


class ResourceB(AbstractResource):
    pass


def test_resolve_resource_using_full_type_field():
    actual = create_resource_from_dict(
        {"type": "au.com.example.abstracts.ResourceA"},
        AbstractResource,
        full_clean=False,
    )

    assert isinstance(actual, ResourceA)


def test_resolve_resource_using_partial_type_field():
    actual = create_resource_from_dict(
        {"type": "ResourceB"},
        AbstractResource,
        full_clean=False,
    )

    assert isinstance(actual, ResourceB)


@pytest.mark.parametrize("resource", (ResourceA, ResourceB))
def test_type_field_is_inherited(resource):
    expected = getmeta(AbstractResource).type_field
    actual = getmeta(resource).type_field

    assert actual == expected


@pytest.mark.parametrize("resource", (ResourceA, ResourceB))
def test_name_space_is_inherited(resource):
    expected = getmeta(AbstractResource).name_space
    actual = getmeta(resource).name_space

    assert actual == expected
